@@ -1,6 +1,6 @@
 #ifndef _JED_CONFIG_H_
 #define _JED_CONFIG_H_
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -103,6 +103,9 @@
 # define HAVE_MEMCHR
 #endif
 
+/* OS/2 probably has gethostname */
+#undef HAVE_GETHOSTNAME
+
 #if !defined(__WATCOMC__) || defined(OS2)
 # define HAS_MOUSE 1
 #endif
@@ -124,9 +127,19 @@
 /* Set this to 1 if the filesystem is case-sensitive */
 #define JED_FILE_PRESERVE_CASE 0
 
+/* Define if you have the vsnprintf, snprintf functions and they return
+ * EOF upon failure.
+ */
+#undef HAVE_VSNPRINTF
+#undef HAVE_SNPRINTF
+
 /* Finally, a reality check... */
 #ifdef REAL_UNIX_SYSTEM
 # undef REAL_UNIX_SYSTEM
 #endif
 
+#ifndef JED
+# define JED
+#endif
+
 #endif				       /* _JED_CONFIG_H_ */
