@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -33,6 +33,7 @@
 #include "ins.h"
 #include "hooks.h"
 #include "replace.h"
+#include "indent.h"
 #if JED_HAS_TTY_MENUS
 # include "menu.h"
 #endif
@@ -44,6 +45,8 @@
 /*}}}*/
 
 void (*X_Define_Keys_Hook)(SLKeyMap_List_Type *);
+int (*X_Set_Abort_Char_Hook) (unsigned char);
+unsigned char Jed_Abort_Char = 7;		       /* ^G */
 
 typedef struct /*{{{*/
 {
@@ -56,6 +59,8 @@ jmp_buf_struct;
 jmp_buf_struct Jump_Buffer, *Jump_Buffer_Ptr;
 
 FVOID_STAR Last_Key_Function;
+FVOID_STAR Jed_This_Key_Function;
+
 int *Repeat_Factor;                    /* repeats a key sequence if non null */
 
 char *Read_This_Character = NULL;      /* alternate keyboard buffer */
@@ -83,6 +88,7 @@ static int redraw (void)
    return 0;
 }
 
+#if HAS_MOUSE
 #ifndef IBMPC_SYSTEM
 static int xterm_mouse_cmd (void)
 {
@@ -123,32 +129,43 @@ static int xterm_mouse_cmd (void)
    return 1;
 }
 #endif
+#endif				       /* HAS_MOUSE */
 
 SLKeymap_Function_Type Jed_Functions[] = /*{{{*/
 {
+   /* Put common cases here to speedup lookup */
+     {"self_insert_cmd", ins_char_cmd},
+     {"next_char_cmd", next_char_cmd},
+     {"next_line_cmd", next_line_cmd},
+     {"previous_char_cmd", previous_char_cmd},
+     {"previous_line_cmd", previous_line_cmd},
+     {"delete_char_cmd", delete_char_cmd},
+     {"beg_of_line", bol},
+     {"eol_cmd", eol_cmd},
+     {"beg_of_buffer", bob},
+     {"end_of_buffer", eob},
+     {"newline_and_indent", newline_and_indent},
+     {"newline", newline_cmd},
+     {"page_down", pagedown_cmd},
+     {"page_up", pageup_cmd},
      {"backward_delete_char", backward_delete_char_cmd},
      {"backward_delete_char_untabify", backward_delete_char_untabify},
      {"beep", jed_beep},
      {"begin_macro", begin_keyboard_macro},
-     {"beg_of_buffer", bob},
-     {"beg_of_line", bol},
      {"center_line", center_line},
      {"copy_region", copy_to_pastebuffer},
-#if defined(__MSDOS__) && !defined(__WIN32__)
+#if (defined(__MSDOS__) && !defined(__WIN32__)) || (defined(__QNX__) && defined(__WATCOMC__))
      {"coreleft", show_memory},
 #endif
-     {"delete_char_cmd", delete_char_cmd},
      {"delete_window", delete_window},
      {"digit_arg", digit_arg},
 /*      {"double_line", double_line}, */
      {"end_macro", end_keyboard_macro},
      {"enlarge_window", enlarge_window},
-     {"end_of_buffer", eob},
-     {"eol_cmd", eol_cmd},
      {"evaluate_cmd", evaluate_cmd},
      {"execute_macro", execute_keyboard_macro},
      {"exchange",exchange_point_mark},
-     {"exit_jed", exit_jed},
+     {"exit_jed", jed_exit_jed_cmd},
      {"exit_mini", exit_minibuffer},
      {"find_file", find_file},
      {"format_paragraph", text_format_paragraph},
@@ -168,26 +185,17 @@ SLKeymap_Function_Type Jed_Functions[] = /*{{{*/
      {"narrow", narrow_to_lines},
      {"narrow_paragraph", narrow_paragraph},
      {"narrow_to_region", narrow_to_region},
-     {"newline", newline_cmd},
-     {"newline_and_indent", newline_and_indent},
-     {"next_char_cmd", next_char_cmd},
-     {"next_line_cmd", next_line_cmd},
      {"one_window", one_window},
      {"other_window", other_window},
-     {"page_down", pagedown_cmd},
-     {"page_up", pageup_cmd},
      {"pop_spot", pop_spot},
-     {"previous_char_cmd", previous_char_cmd},
-     {"previous_line_cmd", previous_line_cmd},
      {"quoted_insert", quoted_insert},
      {"redraw", redraw},
       /* {"replace", replace}, */
-     {"save_buffers", save_some_buffers},
+     {"save_some_buffers", save_some_buffers},
      {"scroll_left", scroll_left},
      {"scroll_right", scroll_right},
        /* {"search_backward", search_backward_cmd},
 	  {"search_forward", search_forward_cmd}, */
-     {"self_insert_cmd", ins_char_cmd},
      {"set_mark_cmd", set_mark_cmd},
      {"split_window", split_window},
      {"switch_to_buffer", get_buffer},
@@ -198,7 +206,7 @@ SLKeymap_Function_Type Jed_Functions[] = /*{{{*/
      {"undo", undo},
      {"widen", widen},
      {"widen_region", widen_region},
-     {"write_buffer", write_buffer},
+     {"write_buffer", jed_save_buffer_as_cmd},
      {"yank", yank},
 #if JED_HAS_TTY_MENUS
      {"select_menubar", jed_select_menu_bar},
@@ -209,13 +217,15 @@ SLKeymap_Function_Type Jed_Functions[] = /*{{{*/
 /*}}}*/
 
 SLKeyMap_List_Type *Global_Map;
-SLKeyMap_List_Type *Mini_Map;
 
 int kbd_quit(void) /*{{{*/
 {
    int sle = SLang_Error;
    SLKeyBoard_Quit = 1;
-   
+#if JED_HAS_TTY_MENUS
+   if (Jed_Menus_Active)
+     jed_exit_menu_bar ();
+#endif
    msg_error("Quit!");
    if (Ignore_User_Abort) SLang_Error = sle;
 #ifdef UNDO_HAS_REDO
@@ -298,7 +308,7 @@ void init_keymaps(void) /*{{{*/
    SLkm_define_key (PC_ALT_F5, (FVOID_STAR) set_mark_cmd, Global_Map);
    SLkm_define_key (PC_SHIFT_F4, (FVOID_STAR) narrow_paragraph, Global_Map);
    SLkm_define_key (PC_SHIFT_F6, (FVOID_STAR) center_line, Global_Map);
-   SLkm_define_key (PC_F7, (FVOID_STAR) exit_jed, Global_Map);
+   SLkm_define_key (PC_F7, (FVOID_STAR) jed_exit_jed_cmd, Global_Map);
 #else
    
    /* give vtxxx arrow keys */
@@ -327,8 +337,8 @@ void init_keymaps(void) /*{{{*/
 #endif
 #endif   
    
-   SLkm_define_key ("'", (FVOID_STAR) text_smart_quote, Global_Map);
-   SLkm_define_key ("\"", (FVOID_STAR) text_smart_quote, Global_Map);
+   /* SLkm_define_key ("'", (FVOID_STAR) text_smart_quote, Global_Map);
+   SLkm_define_key ("\"", (FVOID_STAR) text_smart_quote, Global_Map); */
    SLkm_define_key ("^_", (FVOID_STAR) undo, Global_Map);
    SLkm_define_key ("^?", (FVOID_STAR) backward_delete_char_untabify, Global_Map);
 #ifndef IBMPC_SYSTEM
@@ -346,7 +356,7 @@ void init_keymaps(void) /*{{{*/
    SLkm_define_key ("^K(", (FVOID_STAR) begin_keyboard_macro, Global_Map);
    SLkm_define_key ("^K)", (FVOID_STAR) end_keyboard_macro, Global_Map);
    SLkm_define_key ("^KD", (FVOID_STAR) evaluate_cmd, Global_Map);
-   SLkm_define_key ("^KE", (FVOID_STAR) exit_jed, Global_Map);
+   SLkm_define_key ("^KE", (FVOID_STAR) jed_exit_jed_cmd, Global_Map);
    SLkm_define_key ("^KG", (FVOID_STAR) find_file, Global_Map);
    SLkm_define_key ("^KK", (FVOID_STAR) copy_to_pastebuffer, Global_Map);
    SLkm_define_key ("^KM", (FVOID_STAR) mark_spot, Global_Map);
@@ -358,7 +368,7 @@ void init_keymaps(void) /*{{{*/
    SLkm_define_key ("^K^P", (FVOID_STAR) yank, Global_Map);
 /*     SLang_define_key1("^K^R", (VOID_STAR) replace, SLKEY_F_INTRINSIC, Global_Map); */
    SLkm_define_key ("^K^V", (FVOID_STAR) kill_region, Global_Map);
-   SLkm_define_key ("^K^W", (FVOID_STAR) write_buffer, Global_Map);
+   SLkm_define_key ("^K^W", (FVOID_STAR) jed_save_buffer_as_cmd, Global_Map);
    SLkm_define_key ("^L", (FVOID_STAR) kill_line, Global_Map);
    SLkm_define_key ("^M", (FVOID_STAR) newline_and_indent, Global_Map);
    SLkm_define_key ("^R", (FVOID_STAR) redraw, Global_Map);
@@ -405,82 +415,6 @@ void set_current_kbd_command (char *s) /*{{{*/
 
 /*}}}*/
 
-static int key_interpret(SLang_Key_Type *key) /*{{{*/
-{
-   char *rtc_save;
-   char *str;
-   int ret;
-   static int macro_depth;
-   
-   strcpy (Last_Kbd_Command_String, Current_Kbd_Command_String);
-   
-   if (key->type == SLKEY_F_INTRINSIC)
-     {
-	set_current_kbd_command (Jed_Key_Buffer);
-	Current_Kbd_Command_String[31] = 0;
-	ret = (key->f.f)();
-     }
-   else
-     {
-	str = key->f.s;
-	set_current_kbd_command (str);
-	if (*str == ' ')
-	  {
-	     /* string to insert */
-	     insert_string(str + 1);
-	  }
-	else if ((*str == '@') && (0 != *(str + 1)))
-	  {
-	     int repeat;
-	     
-	     if (Repeat_Factor != NULL)
-	       {
-		  repeat = *Repeat_Factor;
-		  Repeat_Factor = NULL;
-	       }
-	     else repeat = 1;
-	     
-	     if (macro_depth > 10)
-	       {
-		  macro_depth = 0;
-		  msg_error ("Possible runaway macro aborted.");
-		  return 1;
-	       }
-	     
-	     macro_depth++;
-	     rtc_save = Read_This_Character;
-	     
-	     while (repeat-- > 0)
-	       {
-		  Read_This_Character = str + 1;
-		  do
-		    {
-		       do_key();
-		    }
-		  while ((Read_This_Character != NULL)
-			 && (SLKeyBoard_Quit  == 0)
-			 && (SLang_Error == 0));
-		  
-		  if (SLKeyBoard_Quit || SLang_Error)
-		    break;
-	       }
-	     
-	     Read_This_Character = rtc_save;
-	     macro_depth--;
-	  }
-	else if ((*str == '.')
-		 || !SLang_execute_function(str))
-	  SLang_load_string(str);
-	
-	ret = 1;
-     }
-   
-   Last_Key_Function = key->f.f;
-   return(ret);
-}
-
-/*}}}*/
-
 static void update_jed_keybuffer (void) /*{{{*/
 {
    Last_Key_Buffer_Len = (int) (Key_Bufferp - Key_Buffer);
@@ -500,10 +434,125 @@ static void update_jed_keybuffer (void) /*{{{*/
 
 /*}}}*/
 
+static int do_macro_string (char *str, int repeat)
+{
+   static int macro_depth;
+   char *rtc_save;
+   int ret = 0;
 
-#define XKEY(key)  key_interpret((key))
+   if (macro_depth > 10)
+     {
+	macro_depth = 0;
+	msg_error ("Possible runaway macro aborted.");
+	return -1;
+     }
+	     
+   macro_depth++;
+   rtc_save = Read_This_Character;
+	     
+   while (repeat > 0)
+     {
+	repeat--;
 
-int do_key (void) /*{{{*/
+	Read_This_Character = str;
+   
+	while ((SLang_Error == 0)
+	       && (SLKeyBoard_Quit == 0)
+	       && (Read_This_Character != NULL))
+	  jed_do_key();
+	
+	if (Read_This_Character != NULL)
+	  {
+	     ret = -1;
+	     break;
+	  }
+     }
+
+   Read_This_Character = rtc_save;
+   macro_depth--;
+
+   return ret;
+}
+
+static int key_interpret (SLang_Key_Type *key) /*{{{*/
+{
+   char *str;
+   int ret;
+   
+   strcpy (Last_Kbd_Command_String, Current_Kbd_Command_String);
+   Jed_This_Key_Function = key->f.f;
+
+   if (key->type == SLKEY_F_INTRINSIC)
+     {
+	set_current_kbd_command (Jed_Key_Buffer);
+	ret = (key->f.f) ();
+     }
+   else
+     {
+	str = key->f.s;
+	set_current_kbd_command (str);
+
+	if (*str == ' ')
+	  {
+	     /* string to insert */
+	     (void) jed_insert_string(str + 1);
+	  }
+	else if (*str == '@')
+	  {
+	     int repeat;
+
+	     if (Repeat_Factor != NULL)
+	       {
+		  repeat = *Repeat_Factor;
+		  Repeat_Factor = NULL;
+	       }
+	     else repeat = 1;
+
+	     (void) do_macro_string (str + 1, repeat);
+	  }
+	else if ((*str == '.')
+		 || !SLang_execute_function(str))
+	  SLang_load_string(str);
+	
+	ret = 1;
+     }
+   
+   Last_Key_Function = Jed_This_Key_Function;
+   return(ret);
+}
+
+/*}}}*/
+
+static char *find_function_string (FVOID_STAR f) /*{{{*/
+{
+   SLKeymap_Function_Type *fp;
+   
+   if (f == (FVOID_STAR) ins_char_cmd) return "self_insert_cmd";
+   
+   fp = Jed_Functions;
+   while ((fp != NULL) && (fp->name != NULL))
+     {
+	if ((FVOID_STAR) fp->f == f) return fp->name;
+	fp++;
+     }
+   return NULL;
+}
+
+/*}}}*/
+
+static char *lookup_key_function_string (SLang_Key_Type *key)
+{
+   if (key == NULL)
+     return NULL;
+
+   if (key->type != SLKEY_F_INTRINSIC)
+     return key->f.s;
+   
+   return find_function_string (key->f.f);
+}
+
+
+static int do_key (void) /*{{{*/
 {
    SLang_Key_Type *key;
    int repeat;
@@ -519,23 +568,34 @@ int do_key (void) /*{{{*/
      key = jed_menu_do_key ();
    else
 #endif
-     key = SLang_do_key (CBuf->keymap, jed_getkey);
+     {
+	key = SLang_do_key (CBuf->keymap, jed_getkey);
+	if ((IS_MINIBUFFER == 0)
+	    && (key != NULL)
+	    && jed_hook_exists ("_jed_before_key_hooks"))
+	  {
+	     (void) jed_va_run_hooks ("_jed_before_key_hooks", JED_HOOKS_RUN_ALL, 
+				      1, lookup_key_function_string (key));
+	  }
+     }
 
    update_jed_keybuffer ();
    
    if ((key != NULL) && (key->f.f != NULL))
      {
-	if (Repeat_Factor == NULL) return XKEY(key);
+	if (Repeat_Factor == NULL) return key_interpret (key);
 	repeat = *Repeat_Factor;
 	Suspend_Screen_Update = 1;
 	
 	/* some routines may use the repeat factor as a prefix argument */
-	while (repeat-- > 0)
+	while (repeat > 0)
 	  {
-	     if (SLKeyBoard_Quit || SLang_Error ||
-		 (Repeat_Factor == NULL)) break;
-	 /* (!Executing_Keyboard_Macro && (Repeat_Factor == NULL))) break; */
-	     XKEY(key);
+	     repeat--;
+
+	     if (SLKeyBoard_Quit || SLang_Error || (Repeat_Factor == NULL)) 
+	       break;
+
+	     key_interpret (key);
 	  }
 	Repeat_Factor = NULL;
 	
@@ -552,18 +612,34 @@ int do_key (void) /*{{{*/
 
 /*}}}*/
 
+int jed_do_key (void)
+{
+   int ret = do_key ();
+   if (SLang_Error)
+     {
+	jed_abort_keyboard_macro ();
+	Read_This_Character = NULL;
+	Repeat_Factor = NULL;
+     }
+   return ret;
+}
+
+
+
 void do_jed(void) /*{{{*/
 {
    char *mode;
-   Buffer *tthis = CBuf;
+   Buffer *tthis;
+
+   tthis = CBuf;
+   mode = tthis->mode_string;
 
    /* Mark Undo boundary now because this might not be valid later */
    mark_undo_boundary (tthis);
-   mode = CBuf->mode_string;
 
    Repeat_Factor = NULL;
    Replace_Preserve_Case = 0;
-   if (do_key()) JWindow->trashed = 1;
+   if (jed_do_key()) JWindow->trashed = 1;
    
     /* internal editing commands may have selected a different buffer
      * so put it back to the one associated with the window.
@@ -571,7 +647,9 @@ void do_jed(void) /*{{{*/
    if (CBuf != JWindow->buffer)
      {
 	char *name;
-	if (buffer_exists(JWindow->buffer)) name = JWindow->buffer->name; else name = "*scratch*";
+	if (buffer_exists(JWindow->buffer))
+	  name = JWindow->buffer->name; 
+	else name = "*scratch*";
 	switch_to_buffer_cmd(name);
      }
 
@@ -583,23 +661,43 @@ void do_jed(void) /*{{{*/
 
 /*}}}*/
 
-void jed(void) /*{{{*/
+static int run_switch_active_buffer_hooks (char **bnamep)
+{
+   char *buffer_name = *bnamep;
+
+   if (buffer_name == CBuf->name)
+     return 0;
+
+   if (buffer_name == NULL)
+     (void) jed_va_run_hooks ("_jed_switch_active_buffer_hooks", JED_HOOKS_RUN_ALL, 1, "");
+   else
+     {
+	(void) jed_va_run_hooks ("_jed_switch_active_buffer_hooks", JED_HOOKS_RUN_ALL, 1, buffer_name);
+	SLang_free_slstring (buffer_name);
+     }
+   
+   /* This should not fail because CBuf->name is an slstring, and if it
+    * does fail, it is not important.
+    */
+   *bnamep = SLang_create_slstring (CBuf->name);
+   return 0;
+}
+
+
+void jed (void) /*{{{*/
 {
    /* This routine is called from main.  So before actually strting, check
       one more hook to just to make sure  all things are go. */
-   char *startup = ".()jed_startup_hook";
+   char *buffer_name;
 
 #if JED_HAS_TTY_MENUS
    jed_notify_menu_buffer_changed ();
 #endif
-   
-   if (2 == SLang_is_defined(startup + 3))
-     {
-	SLang_run_hooks(startup + 3, 0);
-	/* remove its definition from memory */
-	SLang_load_string (startup);
-     }
-   
+
+   buffer_name = NULL;
+   (void) run_switch_active_buffer_hooks (&buffer_name);
+   (void) jed_va_run_hooks ("_jed_startup_hooks", JED_HOOKS_RUN_ALL, 0);
+   (void) run_switch_active_buffer_hooks (&buffer_name);
    
    Jump_Buffer_Ptr = &Jump_Buffer;
    
@@ -621,9 +719,11 @@ void jed(void) /*{{{*/
    JWindow->trashed = 1;
    update((Line *) NULL, 0, 0, 1);
    Read_This_Character = NULL;
+
    while(1)
      {
 	Suspend_Screen_Update = 0;
+
 	do_jed();
 	if (!SLKeyBoard_Quit && (CBuf->flags & BUFFER_MODIFIED)
 	    && (!Cursor_Motion)) CBuf->hits += 1;
@@ -632,7 +732,10 @@ void jed(void) /*{{{*/
 	     auto_save_buffer(CBuf);
 	     check_buffers();   /* check files on disk to see if they are recent */
 	  }
-	
+	(void) run_switch_active_buffer_hooks (&buffer_name);
+	if (Jed_Menus_Active == 0)
+	  (void) jed_va_run_hooks ("_jed_after_key_hooks", JED_HOOKS_RUN_ALL, 0);
+
 	if ((Last_Key_Function != (FVOID_STAR) ins_char_cmd)
 	    || (JWindow->trashed) || (JWindow != JWindow->next)
 	    || Suspend_Screen_Update ||
@@ -655,7 +758,7 @@ int digit_arg(void) /*{{{*/
    i = 0;
    buf[i++] = (char) SLang_Last_Key_Char;
    
-   /* After do_key (what called this), Key_Bufferp is reset.  However, I want 
+   /* After jed_do_key (what called this), Key_Bufferp is reset.  However, I want 
     * to keep it for echoing subsequent characters.  I restore its previous 
     * value so that echoing will continue. 
     */
@@ -683,7 +786,7 @@ int digit_arg(void) /*{{{*/
    if (Key_Bufferp != Key_Buffer) Key_Bufferp--;
    
    /* Key_Timeout is still active and is only reset after this call. */
-   do_key();
+   jed_do_key();
    return(1);
 }
 
@@ -735,23 +838,6 @@ int which_key (char *f) /*{{{*/
 
 /*}}}*/
 
-static char *find_function_string (FVOID_STAR f) /*{{{*/
-{
-   SLKeymap_Function_Type *fp;
-   
-   fp = Jed_Functions;
-   
-   if (f == (FVOID_STAR) ins_char_cmd) return "self_insert_cmd";
-   
-   while ((fp != NULL) && (fp->name != NULL))
-     {
-	if ((FVOID_STAR) fp->f == f) return fp->name;
-	fp++;
-     }
-   return NULL;
-}
-
-/*}}}*/
 
 static void dump_this_binding (SLang_Key_Type *key) /*{{{*/
 {
@@ -794,20 +880,16 @@ static void dump_this_binding (SLang_Key_Type *key) /*{{{*/
 	     *(str + 1) = ch + '@';
 	     n = 2;
 	  }
-	quick_insert((unsigned char *) str, n);
-	ins(' ');
+	(void) jed_quick_insert((unsigned char *) str, n);
+	(void) jed_ins(' ');
      }
-   ins_char_n_times('\t', 3);
+   (void) jed_ins_char_n_times('\t', 3);
    
-   if (key->type == SLKEY_F_INTRINSIC)
-     {
-   	fun = find_function_string (key->f.f);
-     }
-   else fun = key->f.s;
+   fun = lookup_key_function_string (key);
    
    if (fun == NULL) fun = "** Unknown **";
-   insert_string (fun);
-   newline();
+   (void) jed_insert_string (fun);
+   (void) jed_insert_newline ();
 }
 
 /*}}}*/
@@ -845,36 +927,6 @@ void dump_bindings(char *map) /*{{{*/
 
 /*}}}*/
 
-char *find_key(int *ret) /*{{{*/
-{
-   char *fstr = NULL;
-   SLang_Key_Type *key;
-   
-   *ret = 0;
-   
-   SLang_Key_TimeOut_Flag = 0;
-   Key_Bufferp = Key_Buffer;
-   key = SLang_do_key (CBuf->keymap, jed_getkey);
-   update_jed_keybuffer ();
-   
-   if (key == NULL)
-     {
-	flush_input ();
-	return NULL;
-     }
-
-   if (key->type == SLKEY_F_INTRINSIC)
-     {
-	*ret = 1;
-	fstr = find_function_string (key->f.f);
-     }
-   else fstr = key->f.s;
-   
-   return fstr;
-}
-
-/*}}}*/
-
 void use_keymap(char *name) /*{{{*/
 {
    SLKeyMap_List_Type *map;
@@ -896,25 +948,18 @@ char *what_keymap() /*{{{*/
 
 /*}}}*/
 
-void set_abort_char(int *c) /*{{{*/
+void jed_set_abort_char (int *c) /*{{{*/
 {
    char str[2];
    int i;
-#ifdef __MSDOS_16BIT__
-   char ch, *s, *scan = "@#$%^&*()_+?IQWERTYUIOP[]!!ASDFGHJKL!!!!\\ZXCVBNM";
-   
-   ch = 64 + (char) *c;
-   s = scan; while (*s && (*s != ch)) s++;
-   if (*s == 0) return;
-   Abort_Char = (int) (s - scan) + 0x03;
-#else
-   Abort_Char = *c;
-#endif
-#ifndef MSWINDOWS
-   reset_tty();
-   init_tty();
-#endif
-   str[0] = *c; str[1] = 0;
+   char ch = (char) *c;
+
+   if (X_Set_Abort_Char_Hook != NULL)
+     (void) (*X_Set_Abort_Char_Hook) ((unsigned char) ch);
+
+   Jed_Abort_Char = (int) ch;
+
+   str[0] = ch; str[1] = 0;
    
    for (i = 0; i < SLANG_MAX_KEYMAPS; i++)
      {
@@ -935,7 +980,6 @@ static char **Slang_Flist_Context;
 
 int next_function_list(char *buf) /*{{{*/
 {
-   SLKeymap_Function_Type *tthis = Jed_Functions;
    register char *name;
    char **max;
    
@@ -950,7 +994,7 @@ int next_function_list(char *buf) /*{{{*/
    
    while (1)
      {
-	tthis = Flist_Context;
+	SLKeymap_Function_Type *tthis = Flist_Context;
 	name = tthis->name;
 	if (name == NULL) break;
 	Flist_Context++;
@@ -1033,12 +1077,120 @@ int is_internal(char *f) /*{{{*/
 
 /*}}}*/
 
-void create_keymap (char *name) /*{{{*/
+void jed_copy_keymap_cmd (char *dest, char *src)
 {
-   if (NULL == SLang_create_keymap (name, Global_Map))
+   SLKeyMap_List_Type *ksrc;
+
+   ksrc = SLang_find_keymap (src);
+   if (ksrc == NULL)
      {
-	msg_error ("Unable to create keymap.");
+	jed_verror ("Keymap \"%s\" does not exist",  src);
+	return;
+     }
+   if (NULL != SLang_find_keymap (dest))
+     {
+	jed_verror ("Keymap \"%s\" already exists", dest);
+	return;
+     }
+
+   if (NULL == SLang_create_keymap (dest, ksrc))
+     jed_verror ("Unable to create keymap \"%s\"", dest);
+}
+
+
+typedef struct
+{
+   unsigned char *key;
+   unsigned int len;
+   unsigned int num_read;
+}
+Getkey_Callback_Data_Type;
+
+static Getkey_Callback_Data_Type Getkey_Callback_Data;
+
+static int get_key_function_callback (void)
+{
+   unsigned int i = Getkey_Callback_Data.num_read;
+   
+   if (i == Getkey_Callback_Data.len)
+     return SLANG_GETKEY_ERROR;
+   
+   Getkey_Callback_Data.num_read += 1;
+   return (int) Getkey_Callback_Data.key[i];
+}
+
+   
+int jed_get_key_function (unsigned char *keystr, char **fun)
+{
+   SLang_Key_Type *key;
+   
+   SLang_Key_TimeOut_Flag = 0;
+
+   *fun = NULL;
+   if (keystr == NULL)
+     {	
+	Key_Bufferp = Key_Buffer;
+	key = SLang_do_key (CBuf->keymap, jed_getkey);
+	update_jed_keybuffer ();
+	if (key == NULL)
+	  flush_input ();
+
+	if (SLKeyBoard_Quit && (SLang_Error == USER_BREAK))
+	  {
+	     SLang_Error = 0;
+	     SLKeyBoard_Quit = 0;
+	  }
+     }
+   else
+     {
+	Getkey_Callback_Data.len = (unsigned int) keystr[0];
+	Getkey_Callback_Data.key = keystr + 1;
+	Getkey_Callback_Data.num_read = 0;
+	key = SLang_do_key (CBuf->keymap, get_key_function_callback);
+     }
+   
+   if (key == NULL)
+     return -1;
+   
+   if (key->type == SLKEY_F_INTRINSIC)
+     {
+	*fun = find_function_string (key->f.f);
+	return 1;
+     }
+   else 
+     {
+	*fun = key->f.s;
+	if (key->f.s[0] == '@') return 2;
+	if (key->f.s[0] == ' ') return 3;
+	return 0;
      }
 }
 
 /*}}}*/
+   
+   
+   
+     
+   
+void jed_call_cmd (char *str)
+{
+   SLKeyMap_List_Type *km;
+   int (*fp)(void);
+
+   if (*str == ' ')
+     {
+	jed_insert_string (str + 1);
+	return;
+     }
+   
+   if (*str == '@')
+     {
+	(void) do_macro_string (str + 1, 1);
+	return;
+     }
+   
+   km = CBuf->keymap;
+   if (NULL == (fp = (int (*)(void)) (SLang_find_key_function(str, km))))
+     jed_verror ("Internal function %s does not exist", str);
+   else (void) (*fp)();
+}
