@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -61,6 +61,8 @@
 #include "hooks.h"
 #include "undo.h"
 #include "menu.h"
+#include "userinfo.h"
+#include "indent.h"
 
 #ifdef __WIN32__
 # include "win32.h"
@@ -87,14 +89,15 @@ int MiniBuffer_Active = 0;
 
 char *Jed_Library;
 
-int (*complete_open)(char *);
-int (*complete_next)(char *);
+/* Do not change this value without changing sys_findfirst/next. */
+#define JED_MAX_COMPLETION_LEN JED_MAX_PATH_LEN   
+
+static int (*complete_open)(char *);
+static int (*complete_next)(char *);
 static int Buf_List_Len;
 Buffer *Buf_Context;
 
 static char *Expect_File_Error = "Expecting filename.";
-static char *Keymap_Error = "Unknown_Keymap";
-
 
 static int next_bufflist(char *buf) /*{{{*/
 {
@@ -107,14 +110,14 @@ static int next_bufflist(char *buf) /*{{{*/
 
 	if ((Buf_List_Len == 0)
 #if JED_FILE_PRESERVE_CASE
-	    || (0 == strncmp(buf, tthis->name, Buf_List_Len))
+	    || (0 == strncmp (buf, tthis->name, Buf_List_Len))
 #else
 	    || (0 == jed_case_strncmp (buf, tthis->name, Buf_List_Len))
 #endif
 	    )
 	  {
 	     if (*tthis->name == ' ') continue;   /* internal buffers */
-	     strcpy (buf, tthis->name);
+	     safe_strcpy (buf, tthis->name, JED_MAX_COMPLETION_LEN);
 	     return 1;
 	  }
      }
@@ -216,7 +219,7 @@ int kill_buffer_cmd(char *name) /*{{{*/
 {
    Buffer *this_buf, *kill_buf, *scratch, *buf;
    Window_Type *w;
-   int kill;
+   int do_kill;
    
    if (NULL == (kill_buf = find_buffer(name)))
      {
@@ -242,7 +245,7 @@ int kill_buffer_cmd(char *name) /*{{{*/
      }
 #endif
    
-   kill = 1;
+   do_kill = 1;
    if ((*name != ' ') && (kill_buf->flags & BUFFER_MODIFIED))
      {
 	jed_vmessage (1, "Buffer %s modified. (K)ill, (S)ave first, (A)bort:",
@@ -253,16 +256,16 @@ int kill_buffer_cmd(char *name) /*{{{*/
 	 */
 	switch (my_getkey())
 	  {
-	   case 'k': case 'K': kill = 1; break;
-	   case 's': case 'S': kill = 2; break;
+	   case 'k': case 'K': do_kill = 1; break;
+	   case 's': case 'S': do_kill = 2; break;
 	   default: msg_error("Aborted."); return(0);
 	  }
 	clear_message ();
      }
    
-   if (kill == 2)
+   if (do_kill == 2)
      {
-	write_buffer();
+	jed_save_buffer_as_cmd ();
 	if (SLang_Error)
 	  {
 	     switch_to_buffer(this_buf);
@@ -310,72 +313,109 @@ int kill_buffer_cmd(char *name) /*{{{*/
 
 /*}}}*/
 
-int write_buffer_cmd(char *filestr) /*{{{*/
+int jed_save_buffer_to_file (char *dir, char *file)
 {
-#ifdef VMS
-   char *ff;
-#endif
-   char dir[JED_MAX_PATH_LEN], *f, file[JED_MAX_PATH_LEN];
-   int n;
-   
-#ifdef REAL_UNIX_SYSTEM
-   f = expand_link(filestr);
-#else
-   f = expand_filename(filestr);
-#endif
-   
-   strncpy (file, f, JED_MAX_PATH_LEN); file[JED_MAX_PATH_LEN - 1] = 0;
-   f = extract_file(file);
-   
-   if ((*f == 0) && (*CBuf->file == 0))
+   int status;
+   char *dirfile;
+
+   if ((file == NULL) || (*file == 0))
      {
-	msg_error("Filename Required.");
-	return(0);
-     }
-   
-   n = (int) (f - file);
-   SLMEMCPY((char *) dir, (char *) file, n);
-   dir[n] = 0;
-   
-   if (*f == 0) 
-     {
-	safe_strcpy (file, CBuf->file, sizeof (file));
-	f = file;
+	file = CBuf->file;
+	if (*file == 0)
+	  {	     
+	     msg_error ("Filename Required.");
+	     return -1;
+	  }
      }
 
-   if (file_changed_on_disk (CBuf, dir_file_merge (dir, f)))
+   if (NULL == (dirfile = jed_dir_file_merge (dir, file)))
+     return -1;
+
+   if (file_changed_on_disk (CBuf, dirfile) > 0)
      {
 	if (0 == jed_get_yes_no("File changed on disk.  Save anyway?"))
-	  return 0;
-     }
-	  
-   n = write_file_with_backup (dir, f);
-   if (n >= 0)
-     {
-#ifdef VMS
-	ff = f; while (*ff) if (*ff == ';') *ff = 0; else ff++;
-#endif
-	if (Batch != 2)
 	  {
-	     jed_vmessage (0, "Wrote %d lines to %s%s", n, dir, f);
+	     SLfree (dirfile);
+	     return -1;
 	  }
-	CBuf->flags &= ~BUFFER_MODIFIED;
-	CBuf->flags |= AUTO_SAVE_BUFFER;
-	CBuf->hits = 0;
-#ifdef UNDO_HAS_REDO
-	update_undo_unchanged ();
-#endif
-	visit_file(dir, f);
-	return(1);
      }
-   else
+
+   if (-1 == jed_va_run_hooks ("_jed_save_buffer_before_hooks",
+			       JED_HOOKS_RUN_ALL, 1, dirfile))
      {
-	jed_verror ("Error writing file %s%s", dir, f);
-	return 0;
+	auto_save_buffer (CBuf);
+	SLfree (dirfile);
+	return -1;
      }
+
+   if (((status = write_file_with_backup (dirfile)) >= 0)
+       && (Batch != 2))
+     jed_vmessage (0, "Wrote %d lines to %s", status, dirfile);
+
+   if (status < 0)
+     {
+	jed_verror ("Error writing file %s", dirfile);
+	SLfree (dirfile);
+	return -1;
+     }
+
+	
+   CBuf->flags |= AUTO_SAVE_BUFFER;
+   CBuf->hits = 0;
+
+#ifdef UNDO_HAS_REDO
+   update_undo_unchanged ();
+#endif
+   visit_file (dir, file);
+
+   if (-1 == jed_va_run_hooks ("_jed_save_buffer_after_hooks", JED_HOOKS_RUN_ALL,
+			       1, dirfile))
+     {
+	SLfree (dirfile);
+	return -1;
+     }
+   
+   SLfree (dirfile);
+   return 0;
 }
 
-/*}}}*/
+
+int jed_save_buffer_as (char *filestr)
+{
+   char *dir, *file;
+   int status;
+
+   filestr = jed_expand_filename (filestr);
+   
+   if (-1 == jed_dirfile_to_dir_file (filestr, &dir, &file))
+     {
+	SLfree (filestr);
+	return -1;
+     }
+
+   status = jed_save_buffer_to_file (dir, file);
+   
+   SLfree (filestr);
+   SLfree (dir);
+   SLfree (file);
+
+   if (status == -1)
+     return status;
+
+   /* This function is an intrinsic and is documented to return the 
+    * number of lines written out.
+    * 
+    */
+   return 0;
+}
+
+int jed_save_buffer_cmd (void)
+{
+   if ((CBuf->file == NULL) || (CBuf->file[0] == 0))
+     return jed_save_buffer_as_cmd ();
+   
+   return jed_save_buffer_to_file (CBuf->dir, CBuf->file);
+}
 
 #if defined(__BORLANDC__) && !defined(MSWINDOWS)
 int show_memory() /*{{{*/
@@ -493,6 +533,20 @@ int set_buffer(char *name) /*{{{*/
 int jed_get_mini_response (char *s)
 {
    int ch;
+   int len;
+
+   if (Batch)
+     {
+	if (EOF == fputs (s, stdout))
+	  exit_error ("Failed to write to stdout", 0);
+	
+	return jed_getkey ();
+     }
+	    
+   len = strlen (s);
+   if (len > Jed_Num_Screen_Cols)
+     s += (len - Jed_Num_Screen_Cols);
+
    MiniBuf_Get_Response_String = s;
    update (NULL, 0, 0, 0);
    ch = jed_getkey ();
@@ -531,7 +585,7 @@ int jed_get_y_n (char *question)
      return 1;
    if ((ans == 'n') || (ans == 'N'))
      return 0;
-   if (SLKeyBoard_Quit) 
+   if (SLKeyBoard_Quit || (SLang_Error == SL_USER_BREAK))
      return -1;
    jed_beep ();
    flush_input ();
@@ -548,11 +602,8 @@ int jed_vget_y_n (char *fmt, char *arg)
 	arg = fmt;
 	fmt = "%s";
      }
-   /* FIXME: Priority=low: This is a hack to prevent a possible overflow */
-   if (strlen (arg) + strlen (fmt) >= sizeof (msg))
-     arg = "";
-
-   sprintf (msg, fmt, arg);
+   
+   _SLsnprintf (msg, sizeof (msg), fmt, arg);
    
    return jed_get_y_n (msg);
 }
@@ -597,79 +648,125 @@ int jed_get_yes_no (char *question) /*{{{*/
 
 int find_file_cmd (char *filestr)       /* filestr is const ! */ /*{{{*/
 {
-   char *f, *file, filebuf[JED_MAX_PATH_LEN];
    Buffer *buf;
-   int n;
-   
-#ifdef REAL_UNIX_SYSTEM
-   file = expand_link(filestr);
-#else
-   file = expand_filename(filestr);
-#endif
-   
-   safe_strcpy(filebuf, file, sizeof (filebuf));
-   file = filebuf;
-   f = extract_file(file);
-   if ((*file == 0) || (*f == 0))
+   int status;
+   char *dirfile, *dir, *file;
+   char *name;
+
+   filestr = jed_expand_filename (filestr);
+   if (filestr == NULL)
+     return -1;
+
+   /* If this hook returns 1, then the file is considered to be found. */
+   if (1 == jed_va_run_hooks ("_jed_find_file_before_hooks", JED_HOOKS_RUN_UNTIL_NON_0,
+			      1, filestr))
      {
-	msg_error(Expect_File_Error);
-	return(0);
+	SLfree (filestr);
+	return 0;
      }
+
+   dirfile = jed_expand_link(filestr);
+
+   if (dirfile == NULL)
+     {
+	SLfree (filestr);
+	return -1;
+     }
+
+   if (-1 == jed_dirfile_to_dir_file (filestr, &dir, &file))
+     {
+	SLfree (filestr);
+	SLfree (dirfile);
+	return -1;
+     }
+
+   status = -1;
    
+   if (*file == 0)
+     {
+	jed_verror (Expect_File_Error);
+	goto free_and_return;
+     }
+
    check_buffers ();
    
    /* search for the file in current buffer list */
    
-   if (NULL != (buf = find_file_buffer(file)))
+   if (NULL != (buf = find_file_buffer (dirfile)))
      {
-	if (file_changed_on_disk (buf, file))
+	if ((file_changed_on_disk (buf, dirfile) > 0)
+	    && (1 == jed_get_yes_no("File changed on disk.  Read From Disk")))
 	  {
-	     if (jed_get_yes_no("File changed on disk.  Read From Disk"))
-	       {
-		  if (*Error_Buffer) return(1);
-		  n = (int) buf->linenum;
-		  buf->flags  &= ~BUFFER_MODIFIED;
-		  kill_buffer_cmd(buf->name);
-		  find_file_cmd(file);
-		  goto_line(&n);
-		  return(1);
-	       }
+	     int n = (int) buf->linenum;
+	     mark_buffer_modified (buf, 0, 1);
+	     kill_buffer_cmd (buf->name);
+	     status = find_file_cmd (filestr);
+	     goto_line (&n);
+	     goto free_and_return;
 	  }
-	
+
+	if (SLang_Error)
+	  goto free_and_return;
+
+	status = 1;
 	switch_to_buffer (buf);
-	return(1);
+	goto free_and_return;
      }
-   
-   buf = make_buffer (NULL, file, NULL);
+
+   /* Make a buffer but do not attach it to a file yet. */
+   name = extract_file (filestr);
+
+   buf = make_buffer (name, dir, NULL);
    switch_to_buffer(buf);
-   
-   n = read_file(file);
-   CLine = CBuf->beg;
-   Point = 0;
-   LineNum = 1;
-   if (CLine == NULL) make_line(25);
-   set_file_modes();
-   CBuf->flags |= UNDO_ENABLED;
-   
-   switch(n)
+
+   status = read_file (dirfile);
+	
+   switch (status)
      {
-      case -2: msg_error("File not readable."); break;
+      case -2:
+	jed_verror ("File %s not readable.", dirfile);
+	status = -1;
+	break;
+	     
+      case -1:
+	message ("New file.");
+	status = 0;
+	break;
+	     
       default:
-	if (Batch == 2) break;
-	if (*Message_Buffer) break;   /* autosave warning? */
-	if (n == -1) message("New file.");
-	else
-	  {
-	     jed_vmessage (0, "%d lines read", n);
-	  }
+	jed_vmessage (0, "%d lines read", status);
+	status = 1;
      }
-   
-   SLang_run_hooks("find_file_hook", 1, file);
-   return(1);
+
+   if (buf == CBuf)		       /* check this in case the hook switches buffers */
+     {
+	mark_buffer_modified (CBuf, 0, 0);
+
+	/* Now attach it to a file */
+	visit_file (dir, file);
+	uniquely_name_buffer (CBuf, name);
+
+	(void) bob ();
+	set_file_modes();
+	CBuf->flags |= UNDO_ENABLED;
+     }
+
+   if (-1 == jed_va_run_hooks ("_jed_find_file_after_hooks", JED_HOOKS_RUN_ALL, 0))
+     status = -1;
+
+   free_and_return:
+
+   SLfree (dirfile);
+   SLfree (filestr);
+   SLfree (dir);
+   SLfree (file);
+
+   return status;
 }
 
 /*}}}*/
 
+
 int find_file_in_window(char *file) /*{{{*/
 {
    int ret;
@@ -690,6 +787,13 @@ static void create_minibuffer(void) /*{{{*/
    Window_Type *w;
    MiniBuffer = The_MiniBuffer;
    
+   /* We may get called before the keymap has been set up.  Make sure that
+    * we have a keymap.
+    */
+   if ((The_MiniBuffer->keymap == NULL)
+       || (The_MiniBuffer->keymap == Global_Map))
+     jed_setup_minibuffer_keymap ();
+
    /* I want to make Mini_Window->next = Current Window so that scroll other
       window routines will scroll it. */
    
@@ -697,13 +801,18 @@ static void create_minibuffer(void) /*{{{*/
    do other_window(); while (JWindow->next != w);
    JWindow->next = The_MiniWindow;
    The_MiniWindow->next = w;
-   The_MiniWindow->column = 1;
+   The_MiniWindow->hscroll_column = 1;
    Mini_Info.action_window = w;
    other_window();    /* now in minibuffer window */
    window_buffer(MiniBuffer);
    switch_to_buffer(MiniBuffer);
    MiniBuffer_Active = 1;
    erase_buffer ();
+
+#if JED_HAS_TTY_MENUS
+   jed_notify_menu_buffer_changed ();
+#endif
+
    /* allow kill region to kill to beginning of minibuffer. 
     * Suggested by stefen@uni-paderborn.de */
    push_mark ();
@@ -724,7 +833,7 @@ int exit_minibuffer() /*{{{*/
 	if (Last_Completion_Buffer != NULL)
 	  {
 	     pop_to_buffer (Completion_Buffer);
-	     CBuf->flags &= ~BUFFER_MODIFIED;
+	     mark_buffer_modified (CBuf, 0, 0);
 	     switch_to_buffer_cmd (Last_Completion_Buffer);
 	     kill_buffer_cmd (Completion_Buffer);
 	     touch_window_hard (JWindow, 0);
@@ -748,7 +857,7 @@ int select_minibuffer() /*{{{*/
    w = JWindow;
    while (MiniBuffer != NULL)
      {
-	if (JWindow->top == Jed_Num_Screen_Rows) return(1);
+	if (JWindow->sy+1 == Jed_Num_Screen_Rows) return(1);
 	other_window();
 	if (w == JWindow) exit_error("Internal Error:  no window!", 1);
      }
@@ -790,16 +899,16 @@ static int ledit(void) /*{{{*/
 static char *read_file_from_minibuffer(char *prompt, char *def) /*{{{*/
 {
    int n;
-   char buf[JED_MAX_PATH_LEN];
-   
+   char *file;
+
    complete_open = sys_findfirst;
    complete_next = sys_findnext;
    
    if (*CBuf->dir == 0)
      buffer_filename (CBuf, NULL, CBuf->file);
 
-   strcpy (buf, CBuf->dir);
-   return read_from_minibuffer (prompt, def, buf, &n);
+   file = read_from_minibuffer (prompt, def, CBuf->dir, &n);
+   return file;
 }
 
 /*}}}*/
@@ -851,9 +960,9 @@ static int open_string_list (char *buf) /*{{{*/
 void read_object_with_completion(char *prompt, char *dflt, char *stuff, int *typep) /*{{{*/
 {
    int type = *typep, n;
-   char buf[JED_MAX_PATH_LEN], *tmp;
+   char buf[JED_MAX_COMPLETION_LEN], *tmp;
    char *str = NULL;
-   
+
    *buf = 0;
    if (type == 'f')		       /* file */
      {
@@ -861,7 +970,7 @@ void read_object_with_completion(char *prompt, char *dflt, char *stuff, int *typ
 	complete_next = sys_findnext;
 	if (*CBuf->dir == 0) 
 	  buffer_filename (CBuf, NULL, CBuf->file);
-	strcpy (buf, CBuf->dir);
+	safe_strcpy (buf, CBuf->dir, sizeof (buf));
      }
    else if (type == 'b')	       /* buffer */
      {
@@ -885,7 +994,7 @@ void read_object_with_completion(char *prompt, char *dflt, char *stuff, int *typ
 	complete_open = NULL;
      }
    
-   strcat (buf, stuff);
+   safe_strcat (buf, stuff, sizeof (buf));
    
    if (NULL != (tmp = read_from_minibuffer(prompt, dflt, buf, &n)))
      {
@@ -899,13 +1008,15 @@ void read_object_with_completion(char *prompt, char *dflt, char *stuff, int *typ
 
 /*}}}*/
 
-int insert_file_cmd() /*{{{*/
+int insert_file_cmd (void) /*{{{*/
 {
    char *filebuf, *f, *file;
    
    CHECK_READ_ONLY
      
-     if (NULL == (filebuf = read_file_from_minibuffer("Insert file:", NULL))) return(0);
+     if (NULL == (filebuf = read_file_from_minibuffer("Insert file:", NULL)))
+       return -1;
+
    file = expand_filename(filebuf);
    SLfree (filebuf);
    f = extract_file(file);
@@ -923,35 +1034,35 @@ int insert_file_cmd() /*{{{*/
 
 int find_file (void) /*{{{*/
 {
-   char *tmp, *f;
-   char file [JED_MAX_PATH_LEN];
-   
-   if (NULL == (tmp = read_file_from_minibuffer("Find file:", (char *) NULL))) 
+   char *file, *f;
+   int status;
+
+   if (NULL == (file = read_file_from_minibuffer("Find file:", (char *) NULL))) 
      return 0;
    
-   safe_strcpy (file, tmp, sizeof (file));
-   SLfree(tmp);
-   
    f = extract_file (file);
    if (*f == 0)
      {
-	/* Use buffer filename as default. */
-	safe_strcpy (f, CBuf->file, 
-		     sizeof(file) - (unsigned int) (f - file));
+	char *dirfile = jed_dir_file_merge (file, CBuf->file);
+	SLfree (file);
+	if (dirfile == NULL)
+	  return -1;
+	file = dirfile;
      }
-   find_file_in_window (file);
 
-   return 0;
+   status = find_file_in_window (file);
+   SLfree (file);
+   return status;
 }
 
 /*}}}*/
 
-int write_buffer (void) /*{{{*/
+int jed_save_buffer_as_cmd (void) /*{{{*/
 {
    char *tmp;
    
-   if (NULL == (tmp = read_file_from_minibuffer("Write to file:", (char *) NULL))) return(0);
-   write_buffer_cmd(tmp);
+   if (NULL == (tmp = read_file_from_minibuffer("Save as:", (char *) NULL))) return(0);
+   (void) jed_save_buffer_as (tmp);
    SLfree(tmp);
    return(1);
 }
@@ -1013,7 +1124,10 @@ int kill_buffer (void) /*{{{*/
 	if ((b != NULL) && (b->subprocess))
 	  {
 	     if (0 == jed_get_yes_no("Buffer has a subprocess attached.  Delete anyway"))
-	       return 0;
+	       {
+		  SLfree(tmp);
+		  return 0;
+	       }
 	  }
 #endif
 	kill_buffer_cmd(tmp);
@@ -1032,13 +1146,6 @@ int evaluate_cmd() /*{{{*/
 
 /*}}}*/
 
-void insert_string(char *s) /*{{{*/
-{
-   CHECK_READ_ONLY_VOID
-     ins_chars((unsigned char *) s, strlen(s));
-}
-
-/*}}}*/
 
 /* This is weird, Ultrix cc will not compile if set_key comes before unset_key */
 void unset_key(char *key) /*{{{*/
@@ -1063,7 +1170,7 @@ void unset_key_in_keymap(char *key, char *map) /*{{{*/
    
    if (NULL == (kmap = SLang_find_keymap(map)))
      {
-	msg_error(Keymap_Error);
+	jed_verror ("Unknown keymap: %s", map);
 	return;
      }
    
@@ -1085,7 +1192,7 @@ void set_key_in_keymap(char *f, char *key, char *map) /*{{{*/
    
    if (NULL == (kmap = SLang_find_keymap(map)))
      {
-	msg_error(Keymap_Error);
+	jed_verror ("Unknown keymap: %s", map);
 	return;
      }
    
@@ -1128,7 +1235,7 @@ char *pop_to_buffer(char *name) /*{{{*/
    /* find a window to use */
    do
      {
-	if (w->top != Jed_Num_Screen_Rows)
+	if (w->sy + 1 != Jed_Num_Screen_Rows)
 	  {
 	     if (action != NULL)
 	       {
@@ -1169,7 +1276,7 @@ char *pop_to_buffer(char *name) /*{{{*/
 	     other_window();
 	  }
 	while (JWindow->buffer != w->buffer);
-	JWindow->column = 1;
+	JWindow->hscroll_column = 1;
      }
    
    bname = CBuf->name;
@@ -1205,7 +1312,7 @@ int num_windows() /*{{{*/
 void flush_message(char *m) /*{{{*/
 {
    message(m);
-   if (Batch) return;
+   if (Batch || (JWindow == NULL)) return;
    do_dialog(Message_Buffer);
    SLsmg_gotorc (Jed_Num_Screen_Rows - 1, strlen(Message_Buffer));
    *Message_Buffer = 0;
@@ -1310,7 +1417,7 @@ int pipe_region(char *cmd) /*{{{*/
 #endif
 
 
-static VFILE *jed_open_lib_file (char *file) /*{{{*/
+static VFILE *jed_open_lib_file (char *file, char **dirfile) /*{{{*/
 {
    char libfsl[JED_MAX_PATH_LEN], libfslc[JED_MAX_PATH_LEN];
    char *lib, *type, *libf;
@@ -1321,18 +1428,22 @@ static VFILE *jed_open_lib_file (char *file) /*{{{*/
    lib = Jed_Library;
 
    /* If file begins with ./,then read it from the current directory */
-   if ((*file == '.')
-       && ((file[1] == '/')
+   if ((0 == strncmp (file, "./", 2))
 #ifdef IBMPC_SYSTEM
-	   || (file[1] == '\\')
+       || (0 == strncmp (file, ".\\", 2))
+       || (0 == strncmp (file, "..\\", 3))
 #endif
-       ))
-     lib = "";
+       || (0 == strncmp (file, "../", 3)))
+     {
+	if (NULL == (lib = get_cwd ()))
+	  lib = "";
+     }
    else if ((lib == NULL) || (*lib == 0))
      exit_error("JED_ROOT environment variable needs set.", 0);
 
    if ((NULL != (type = file_type(file)))
-       && (*type == 0))
+       && ((*type == 0) 
+	   || ((0 != strcmp (type, "sl")) && (0 != strcmp (type, "slc")))))
      type = NULL;
 
    n = 0;
@@ -1342,11 +1453,12 @@ static VFILE *jed_open_lib_file (char *file) /*{{{*/
 
 	fixup_dir(libfsl);
 	safe_strcat (libfsl, file, sizeof (libfsl));
-	strcpy (libfsl, expand_filename(libfsl));
+	safe_strcpy (libfsl, expand_filename(libfsl), sizeof (libfsl));
 
 	libf = libfsl;
 
-	if (NULL != (vp = vopen(libf, VFILE_BUF_SIZE, VFILE_TEXT)))
+	if ((1 == file_status (libf))
+	    && (NULL != (vp = vopen(libf, VFILE_BUF_SIZE, VFILE_TEXT))))
 	  break;
 
 	if (type == NULL)
@@ -1362,21 +1474,28 @@ static VFILE *jed_open_lib_file (char *file) /*{{{*/
 #endif
 	     
 	     /* Try .sl and .slc */
-	     strcat (libfsl, ".sl");
-	     strcpy (libfslc, libfsl);
-	     strcat (libfslc, "c");
+	     safe_strcat (libfsl, ".sl", sizeof (libfsl));
+	     safe_strcpy (libfslc, libfsl, sizeof (libfslc));
+	     safe_strcat (libfslc, "c", sizeof (libfslc));
 
 	     if (file_time_cmp(libfslc, libfsl) >= 0)
 	       libf = libfslc;
 
-	     if (NULL != (vp = vopen(libf, VFILE_BUF_SIZE, VFILE_TEXT)))
+	     if ((1 == file_status (libf))
+		 && (NULL != (vp = vopen(libf, VFILE_BUF_SIZE, VFILE_TEXT))))
 	       break;
 	  }
      }
    
-   if (Batch != 2)
+   if (NULL == (*dirfile = SLang_create_slstring (libf)))
+     {
+	vclose (vp);
+	return NULL;
+     }
+
+   if (0 == Jed_Load_Quietly)
      jed_vmessage (1, "loading %s", libf);
-   
+
    return (vp);
 }
 
@@ -1403,16 +1522,21 @@ int jed_load_file (char *file)
    SLang_Load_Type *x;
    int ret;
 
-   if (NULL == (x = SLallocate_load_type (file)))
-     return -1;
-
-   if (NULL == (vp = jed_open_lib_file (file)))
+   if (NULL == (vp = jed_open_lib_file (file, &file)))
      {
 	SLang_verror (SL_OBJ_NOPEN, "Unable to load %s", file);
-	SLdeallocate_load_type (x);
 	return -1;
      }
+
+   x = SLallocate_load_type (file);
+   SLang_free_slstring (file);
    
+   if (x == NULL)
+     {
+	vclose (vp);
+	return -1;
+     }
+
    x->client_data = (VOID_STAR) vp;
    x->read = jed_read_from_file;
    ret = SLang_load_object (x);
@@ -1464,11 +1588,21 @@ void load_buffer (void) /*{{{*/
    Buffer *cbuf = CBuf;
    int flags = CBuf->flags;
    Line *l, *lwant;
+   char *file;
 
-   /* FIXME!!! This should be changed to use the filename if present.  
-    * Otherwise, static objects will not get put in the proper table.
-    */
-   if (NULL == (x = SLallocate_load_type (cbuf->name)))
+   if (cbuf->file[0] == 0)
+     file = cbuf->name;
+   else
+     file = cbuf->file;
+   
+   file = jed_dir_file_merge (CBuf->dir, file);
+   if (file == NULL)
+     return;
+	
+   x = SLallocate_load_type (file);
+   SLfree (file);
+   
+   if (x == NULL)
      return;
 
    x->read = jed_read_from_buffer;
@@ -1494,40 +1628,20 @@ void load_buffer (void) /*{{{*/
    pop_to_buffer (cbuf->name);
    lwant = client_data.l;
 
-   if (lwant != NULL)
+   if (lwant == NULL)
      {
-	bob();
-	while (1)
-	  {
-	     l = CLine->next;
-	     if ((l == NULL) || (l == lwant)) break;
-	     CLine = l; LineNum++;
-	  }
-	(void) skip_whitespace();
+	eob();
+	return;
      }
-}
-
-/*}}}*/
-
-
-void get_key_function() /*{{{*/
-{
-   char *s;
-   int kind;
    
-   s = find_key(&kind);
-   if (s != NULL)
+   bob();
+   while (1)
      {
-	if (SLKeyBoard_Quit && (SLang_Error == USER_BREAK))
-	  {
-	     SLang_Error = 0;
-	     SLKeyBoard_Quit = 0;
-	     /* s = "kbd_quit"; */
-	  }
-	SLang_push_integer(kind);
+	l = CLine->next;
+	if ((l == NULL) || (l == lwant)) break;
+	CLine = l; LineNum++;
      }
-   else s = "";
-   SLang_push_string(s);
+   (void) skip_whitespace();
 }
 
 /*}}}*/
@@ -1570,13 +1684,13 @@ int mini_complete (void) /*{{{*/
 	     safe_strcpy(buf, prev, sizeof (buf));
 	     flag = (*complete_open)(buf);
 	  }
-	strcpy(last, buf);
+	safe_strcpy(last, buf, sizeof (last));
 	n = -1;
      }
    else
      {
 	n = 0;
-	strcpy(prev, buf);  /* save this search context */
+	safe_strcpy (prev, buf, sizeof (prev));  /* save this search context */
      }
    
    if (!n)
@@ -1606,7 +1720,7 @@ int mini_complete (void) /*{{{*/
 		  return 1;
 	       }
 	     safe_strcpy (last, tmp, sizeof (last));
-	     strcpy (prev, last);
+	     safe_strcpy (prev, last, sizeof (prev));
 	     SLang_free_slstring (tmp);
 	     flag = 0;		       /* So we do not call complete_next */
 	     n = -1;
@@ -1617,7 +1731,7 @@ int mini_complete (void) /*{{{*/
    
    if (!n && flag)
      {
-	strcpy(last, buf);
+	safe_strcpy(last, buf, sizeof (last));
 	
 	/* This do loop tests all values from complete_next and returns the
 	   smallest length match of initial characters of buf */
@@ -1628,7 +1742,7 @@ int mini_complete (void) /*{{{*/
 		  set_buffer (Completion_Buffer);
 		  erase_buffer ();
 		  CBuf->flags |= BURIED_BUFFER;
-		  insert_string ("!!! Use Page Up/Down keys to scroll this window. !!!\n");
+		  (void) jed_insert_string ("!!! Use Page Up/Down keys to scroll this window. !!!\n");
 	       }
 	     
 	     n++;
@@ -1636,13 +1750,17 @@ int mini_complete (void) /*{{{*/
 	     pb = buf;
 
 #if !JED_FILE_PRESERVE_CASE
- 	     if (complete_open == open_bufflist)
-	       while (*pl && (UPPER_CASE(*pl) == UPPER_CASE(*pb))) 
+ 	     if ((complete_open == open_bufflist)
+#if !JED_FILENAME_CASE_SENSITIVE
+		 || (complete_open == sys_findfirst)
+#endif
+		 )
+	       while (*pl && (UPPER_CASE(*pl) == UPPER_CASE(*pb)))
 		 {
 		    pl++;
 		    pb++;
 		 }
- 	     else  /* next statement */
+	     else  /* next statement */
 #endif
 	       while (*pl && (*pl == *pb)) 
 	       {
@@ -1656,7 +1774,7 @@ int mini_complete (void) /*{{{*/
 	     if (last_key_char == '\t')
 	       {
 		  while (*pb) pb++;
-		  quick_insert ((unsigned char *)buf, (int) (pb - buf));
+		  (void) jed_quick_insert ((unsigned char *)buf, (int) (pb - buf));
 		  newline ();
 	       }
 	  }
@@ -1669,7 +1787,7 @@ int mini_complete (void) /*{{{*/
 	 */
 	if (complete_open == open_bufflist) 
 	  {
-	     strcpy(buf, last);
+	     safe_strcpy(buf, last, sizeof (buf));
 	     (*complete_open)(buf);
 	     do 
 	       {
@@ -1695,8 +1813,7 @@ int mini_complete (void) /*{{{*/
    if (n)
      {
 	erase_buffer();
-	/* strcpy(last, buf); */
-	insert_string(last);
+	(void) jed_insert_string(last);
 	if ((n == 1) && ((long) Last_Key_Function == (long) mini_complete))
 	  message("[Sole Completion.]");
      }
@@ -1715,20 +1832,6 @@ int what_char() /*{{{*/
 
 /*}}}*/
 
-void call_cmd(char *str) /*{{{*/
-{
-   SLKeyMap_List_Type *km;
-   int (*fp)(void);
-   
-   km = CBuf->keymap;
-   if (NULL == (fp = (int (*)(void)) (SLang_find_key_function(str, km))))
-     {
-	msg_error("Function does not exist!");
-     }
-   else (void) (*fp)();
-}
-
-/*}}}*/
 
 void copy_region_cmd(char *name) /*{{{*/
 {
@@ -1866,7 +1969,7 @@ int dup_mark(void) /*{{{*/
    if (CBuf->marks == NULL) return(0);
    
    push_spot();
-   goto_mark(CBuf->marks);
+   jed_goto_mark(CBuf->marks);
    push_mark();
    pop_spot();
    return(1);
@@ -1887,43 +1990,6 @@ void mini_read(char *prompt, char *def, char *stuff) /*{{{*/
 
 /*}}}*/
 
-void get_buffer_info(void) /*{{{*/
-{
-   SLang_push_string(CBuf->file);
-   SLang_push_string(CBuf->dir);
-   SLang_push_string(CBuf->name);
-   SLang_push_integer(CBuf->flags);
-}
-
-/*}}}*/
-
-void set_buffer_info(char *file, char *dir, char *name, int *flags) /*{{{*/
-{
-   char dirbuf [JED_MAX_PATH_LEN + 1];
-   
-   safe_strcpy (dirbuf, dir, sizeof (dirbuf) - 2);
-   fixup_dir (dirbuf);
-   dir = SLang_create_slstring (expand_filename (dirbuf));
-   file = SLang_create_slstring (file);
-   name = SLang_create_slstring (name);
-   
-   if (SLang_Error)
-     {
-	/* SLang_free_slstring ignores NULL */
-	SLang_free_slstring (dir);
-	SLang_free_slstring (file);
-	SLang_free_slstring (name);
-	return;
-     }
-   
-   CBuf->file = file;
-   CBuf->name = name;
-   CBuf->dir = dir;
-   CBuf->flags = *flags;
-}
-
-/*}}}*/
-
 void make_buffer_list(void) /*{{{*/
 {
    int n = 0;
@@ -1948,9 +2014,11 @@ int window_size_intrinsic(int *what) /*{{{*/
    register int n = 0;
    switch (*what)
      {
+      case 'x': n = JWindow->sx; break;
+      case 'y': n = JWindow->sy; break;
       case 'r': n = JWindow->rows; break;
-      case 'c': n = JWindow->column; break;
-      case 't': n = JWindow->top; break;
+      case 'c': n = JWindow->hscroll_column; break;
+      case 't': n = JWindow->sy + 1; break;
       case 'w': n = JWindow->width; break;
       default: SLang_Error = SL_UNKNOWN_ERROR;
      }
@@ -1959,11 +2027,12 @@ int window_size_intrinsic(int *what) /*{{{*/
 
 /*}}}*/
 
-
+/* This has been moved to slang */
+#if 0
 /* Given a file name with wildcards return expanded list to S-Lang stack
  *  with number.  This does NOT work on unix with wild cards.  Instead the 
  *  expansion is file* (completion) */
-int expand_wildcards(char *file) /*{{{*/
+int expand_wildcards (char *file) /*{{{*/
 {
    char buf[JED_MAX_PATH_LEN];
    int n = 0;
@@ -1983,7 +2052,7 @@ int expand_wildcards(char *file) /*{{{*/
 }
 
 /*}}}*/
-
+#endif
 static void jed_traceback(char *s) /*{{{*/
 {
    char *n;
@@ -1992,7 +2061,7 @@ static void jed_traceback(char *s) /*{{{*/
 	n = CBuf->name;
 	set_buffer("*traceback*");
 	eob();
-	insert_string(s);
+	(void) jed_insert_string(s);
 	set_buffer(n);
      }
    else fprintf(stderr, s);
@@ -2123,17 +2192,63 @@ static void slang_exit_error_handler (char *fmt, va_list ap)
 {
    char buf [2048];
    
-   vsprintf (buf, fmt, ap);
+   _SLvsnprintf (buf, sizeof (buf), fmt, ap);
    exit_error (buf, 0);
 }
 
 static void vmsg_hook (char *fmt, va_list ap)
 {
    char buf [2048];
-   vsprintf (buf, fmt, ap);
+   _SLvsnprintf (buf, sizeof (buf), fmt, ap);
    message (buf);
 }
 
+void jed_setup_minibuffer_keymap (void)
+{
+   SLKeyMap_List_Type *m;
+
+   if (NULL == (m = SLang_find_keymap ("Mini_Map")))
+     {
+	if (NULL == (m = SLang_create_keymap("Mini_Map", Global_Map))) 
+	  exit_error("Unable to create minibuffer keymap", 0);
+
+	SLang_undefine_key("\r", m);
+	SLkm_define_key ("\r", (FVOID_STAR) exit_minibuffer, m);
+	SLkm_define_key ("\t", (FVOID_STAR) mini_complete, m);
+	SLkm_define_key (" ", (FVOID_STAR) mini_complete, m);
+     }
+   The_MiniBuffer->keymap = m;
+}
+
+
+#if SLANG_VERSION < 10405
+/* Work around a slang bug triggered by jed. Apps which use slang's 
+ * load_file function do not have this problem.
+ */
+extern int (*_SLprep_eval_hook) (char *);
+static int bug_fixed_prep_eval_expr (char *expr)
+{
+   int ret;
+   char *end;
+
+   end = strchr (expr, '\n');
+   if (end == NULL)
+     end = expr + strlen (expr);
+   expr = SLmake_nstring (expr, (unsigned int) (end - expr));
+   if (expr == NULL)
+     return -1;
+	
+   if ((0 != SLang_load_string (expr))
+       || (-1 == SLang_pop_integer (&ret)))
+     ret = -1;
+   else
+     ret = (ret != 0);
+
+   SLfree (expr);
+   return ret;
+}
+#endif
+
 void init_minibuffer() /*{{{*/
 {
    Buffer *tmp;
@@ -2145,7 +2260,7 @@ void init_minibuffer() /*{{{*/
    /* do some initializing */
    switch_to_buffer(The_MiniBuffer);
    remake_line(132);
-   The_MiniWindow = create_window(Jed_Num_Screen_Rows, 1, 1, Jed_Num_Screen_Cols);
+   The_MiniWindow = create_window(Jed_Num_Screen_Rows-1, 0, 1, 1, Jed_Num_Screen_Cols);
    The_MiniWindow->buffer = CBuf;
    Buffer_Local.tab = 0;
    switch_to_buffer(tmp);
@@ -2160,6 +2275,10 @@ void init_minibuffer() /*{{{*/
    SLang_Enter_Function = enter_function;
    SLang_Exit_Function = exit_function;
 #endif
+#if SLANG_VERSION < 10405
+   _SLprep_eval_hook = bug_fixed_prep_eval_expr;
+#endif
+
    if ((-1 == SLang_init_slang ())
 #ifndef SIXTEEN_BIT_SYSTEM
        || (-1 == SLang_init_slmath ())
@@ -2167,11 +2286,18 @@ void init_minibuffer() /*{{{*/
        || (-1 == SLang_init_posix_process ())
        || (-1 == SLang_init_posix_dir ())
        || (-1 == SLang_init_slassoc ())
+       || (-1 == SLang_init_stdio ())
+       || (-1 == SLang_init_posix_io ())
+       || (-1 == SLang_init_ospath ())
+       /* || (-1 == SLang_init_import ()) */
        || (-1 == init_jed_intrinsics ())
+       || (-1 == jed_init_userinfo ())
 #if JED_HAS_TTY_MENUS
        || (-1 == jed_init_menus ())
 #endif
+       || (-1 == jed_init_syntax ())
        || (-1 == register_jed_classes ())
+       || (-1 == jed_init_user_hooks ())
        || (-1 == run_init_slang_hooks ()))
      {
 	exit_error("Unable to initialize S-Lang!", 0);
