@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -88,6 +88,9 @@ extern unsigned _stklen = 10000U;
 #include "text.h"
 #include "hooks.h"
 #include "paste.h"
+#include "version.h"
+#include "indent.h"
+#include "colors.h"
 
 #ifdef __WIN32__
 # include "win32.h"
@@ -95,6 +98,8 @@ extern unsigned _stklen = 10000U;
 /*}}}*/
 
 int Jed_Secure_Mode;
+int Jed_Load_Quietly = 0;
+
 char Jed_Root_Dir [JED_MAX_PATH_LEN];
 
 typedef struct /*{{{*/
@@ -114,9 +119,9 @@ char *get_cwd (void) /*{{{*/
    
 #ifdef HAVE_GETCWD
 # if defined (__EMX__)
-   c = _getcwd2(cwd, 254);		       /* includes drive specifier */
+   c = _getcwd2(cwd, sizeof(cwd));		       /* includes drive specifier */
 # else
-   c = getcwd(cwd, 254);		       /* djggp includes drive specifier */
+   c = getcwd(cwd, sizeof(cwd));		       /* djggp includes drive specifier */
 # endif
 #else
    c = (char *) getwd(cwd);
@@ -134,8 +139,13 @@ char *get_cwd (void) /*{{{*/
 	    || (-1 == stat (".", &st2))
 	    || (st1.st_dev != st2.st_dev)
 	    || (st1.st_ino != st2.st_ino))
-	  exit_error ("Unable to get cwd.", 0);
-	
+	  {
+	     jed_vmessage (1, "Unable to get CWD.  Assuming /");
+	     if (-1 == chdir ("/"))
+	       exit_error ("Unable to change directory to /", 0);
+	     c = "/";
+	  }
+
 	strncpy (cwd, c, sizeof (cwd));
 	cwd [sizeof (cwd) - 1] = 0;
 #endif
@@ -156,6 +166,8 @@ static void set_jed_root (char *pgm)
    char *jr, *jl;
    char jed_lib [JED_MAX_PATH_LEN + 1];
 
+   *jed_lib = 0;
+
 #ifdef VMS
    jr = "JED_ROOT:";
 #else
@@ -207,12 +219,18 @@ void (*X_Init_Global_Structures_Hook)(void);
 
 int Stdin_Is_TTY;
 
+int _Jed_Startup_Argc;
+char **_Jed_Startup_Argv;
 
 static int main_initialize (int argc, char **argv)
 {
    char *cmd_hook = ".()command_line_hook";
    int i, fd;
-   
+   char *script_file;
+
+   _Jed_Startup_Argv = argv;
+   _Jed_Startup_Argc = argc;
+
 #ifdef __EMX__
    _response(&argc, &argv);
    _wildcard(&argc, &argv);
@@ -232,7 +250,7 @@ static int main_initialize (int argc, char **argv)
 #ifdef __MSDOS__
    define_word("0-9a-zA-Z\200-\232\240-\245\341-\353");
 #else
-   define_word("0-9a-zA-Z\277-\326\330-\337\340-\366\370-\377");
+   define_word("0-9a-zA-Z\300-\326\330-\366\370-\377");
 #endif
 
 #if JED_HAS_DISPLAY_TABLE
@@ -256,27 +274,61 @@ static int main_initialize (int argc, char **argv)
 	argc -= i;
 	argv += i;
      }
-      
-   if (argc > 1)
+
+   /* If jed-script is a symbolic link to jed, then treat this as equivalent
+    * to jed -script args...
+    */
+   script_file = NULL;
+   if (0 == strcmp (extract_file (argv[0]), "jed-script"))
      {
-	if (!strcmp(argv[1], "-batch"))
+	Batch = 2;
+	SLang_Traceback = 1;
+	Jed_Load_Quietly = 1;
+	if (argc > 1)
+	  script_file = argv[1];
+	else
+	  {
+	     fprintf (stderr, "Usage: jed-script SCRIPT_FILE [args...]\n");
+	     return -1;
+	  }
+	argv++;			       /* make argv[0] the name of the script */
+	argc--;
+     }
+   else if (argc > 1)
+     {
+	char *argv1 = argv[1];
+
+	/* Allow, e.g., --batch */
+	if ((argv1[0] == '-') && (argv1[1] == '-'))
+	  argv1++;
+
+	if (!strcmp(argv1, "-batch"))
 	  {
 	     Batch = 1;
 	     SLang_Traceback = 1;
 	  }
-	else if (!strcmp(argv[1], "-script"))
+	else if (!strcmp(argv1, "-script"))
 	  {
 	     Batch = 2;
 	     SLang_Traceback = 1;
+	     Jed_Load_Quietly = 1;
+	     if (argc > 2) 
+	       script_file = argv[2];
+	  }
+	else if (!strcmp (argv1, "-help"))
+	  {
+	     Batch = 1;
+	     SLang_Traceback = 1;
+	     Jed_Load_Quietly = 1;
 	  }
      }
-   
+
    if (Batch == 0)
      (*tt_get_terminfo) ();
 
    fd = 2;			       /* 2 is stderr, assume it is ok */
 
-#ifdef REAL_UNIX_SYSTEM
+#if defined(REAL_UNIX_SYSTEM) && !defined(__APPLE__)
    if (Batch == 0)
      {
 	int dev_tty_fd;
@@ -312,6 +364,7 @@ static int main_initialize (int argc, char **argv)
      exit_error ("Unable to initialize tty.", 0);
    
    jed_init_display ();         /* sets up virtual screen */
+
    (*tt_set_mono) (JMESSAGE_COLOR, NULL, 0);
    (*tt_set_mono) (JERROR_COLOR, NULL, SLTT_BOLD_MASK);
    (*tt_set_mono) (JDOLLAR_COLOR, NULL, SLTT_BOLD_MASK);
@@ -320,13 +373,14 @@ static int main_initialize (int argc, char **argv)
    (*tt_set_mono) (JMENU_CHAR_COLOR, NULL, SLTT_BOLD_MASK);
    (*tt_set_mono) (JMENU_SELECTION_COLOR, NULL, SLTT_REV_MASK|SLTT_BOLD_MASK);
    (*tt_set_mono) (JMENU_POPUP_COLOR, NULL, 0);
+   (*tt_set_mono) (JMENU_SELECTED_CHAR_COLOR, NULL, SLTT_BOLD_MASK);
+ 
+   (void) jed_set_color (JMENU_SHADOW_COLOR, "blue", "black");
 
-   (*tt_set_color) (JMENU_SHADOW_COLOR, NULL, "blue", "black");
-
-   (*tt_set_color) (JMENU_POPUP_COLOR, NULL, "lightgray", "blue");
+   (void) jed_set_color (JMENU_POPUP_COLOR, "lightgray", "blue");
 
    (*tt_set_mono) (JMENU_SHADOW_COLOR, NULL, 0);
-   (*tt_set_color) (JMENU_SHADOW_COLOR, NULL, "blue", "black");
+   (void) jed_set_color (JMENU_SHADOW_COLOR, "blue", "black");
 
    init_syntax_tables ();
 
@@ -343,10 +397,14 @@ static int main_initialize (int argc, char **argv)
 
    set_file_modes();
    
-   /* what if someone pipes something to jed, allow it */
+   /* what if someone pipes something to jed, allow it unless if is a
+    * jed script.
+    */
    
    if (Stdin_Is_TTY == 0) Stdin_Is_TTY = isatty (0);
-   if (Stdin_Is_TTY == 0)   /* 1 if stdin is a terminal, 0 otherwise */
+   if ((Stdin_Is_TTY == 0)
+       && (Batch != 2))
+     /* 1 if stdin is a terminal, 0 otherwise */
      {
 	set_buffer("*stdin*");
 	read_file_pointer(fileno(stdin));
@@ -362,8 +420,7 @@ static int main_initialize (int argc, char **argv)
 #if !defined(IBMPC_SYSTEM) && !defined(VMS)
    init_signals();
 #endif
-   
-   set_jed_root (argv[0]);
+
    (void) SLang_set_argc_argv (argc, argv);
 
    if ((0 != jed_load_file("site"))
@@ -374,8 +431,11 @@ static int main_initialize (int argc, char **argv)
    
    if (Batch == 2)
      {
-	if (argc > 2)
-	  return jed_load_file (argv[2]);
+	if (0 != jed_load_file (script_file))
+	  return -1;
+
+	if (-1 == SLang_run_hooks ("jedscript_main", 0))
+	  return -1;
      }
    else
      {     
@@ -401,38 +461,31 @@ static int main_initialize (int argc, char **argv)
      return -1;
 
    /* after we have possible loaded key definitions, we can fix up
-    the minibuffer map. This way user definitions are used. */
-   
-   if (NULL == (Mini_Map = SLang_create_keymap("Mini_Map", Global_Map))) exit_error("main: malloc error", 0);
-   SLang_undefine_key("^M", Mini_Map);
-   SLkm_define_key ("^M", (FVOID_STAR) exit_minibuffer, Mini_Map);
-   SLkm_define_key ("^I", (FVOID_STAR) mini_complete, Mini_Map);
-   SLkm_define_key (" ", (FVOID_STAR) mini_complete, Mini_Map);
-   The_MiniBuffer->keymap = Mini_Map;
-   
+    * the minibuffer map. This way user definitions are used. 
+    */
+   jed_setup_minibuffer_keymap ();
+
    return 0;
 }
 
-   
 int main(int argc, char **argv) /*{{{*/
 {
    if (SLang_Version < SLANG_VERSION)
      {
 	fprintf (stderr, "***Warning: Executable compiled against S-Lang %d but linked to %d\n",
-		 SLang_Version, SLANG_VERSION);
+		 SLANG_VERSION, SLang_Version);
 	fflush (stderr);
 	sleep (2);
      }
 
+   set_jed_root (argv[0]);
+
    if (argc > 1)
      {
 	if (!strcmp (argv[1], "--version"))
 	  {
-	     fprintf (stdout, "jed version: %s\n", Jed_Version_String);
-	     fprintf (stdout, "S-Lang version: %s\n", SLang_Version_String);
-	     if (SLang_Version != SLANG_VERSION)
-	       fprintf (stdout, "*** Compiled against S-Lang %d but linked to %d\n",
-			SLang_Version, SLANG_VERSION);
+	     jed_show_version (stdout);
+	     fprintf (stdout, "\nUsing JED_ROOT=%s\n", Jed_Root_Dir);
 	     exit (0);
 	  }
      }
@@ -450,7 +503,7 @@ int main(int argc, char **argv) /*{{{*/
      {
 	jed_reset_display();
 	reset_tty();
-	return 1;
+	return SLang_Error;
      }
 
    /* edit_loop */
@@ -460,7 +513,7 @@ int main(int argc, char **argv) /*{{{*/
    jed_reset_display();
    reset_tty();
 
-   return 0;
+   return SLang_Error;
 }
 
 /*}}}*/
