@@ -1,5 +1,5 @@
 /* Drop-down menus */
-/* Copyright (c) 1999 John E. Davis
+/* Copyright (c) 1999, 2000, 2002 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -23,12 +23,9 @@
 #include "sysdep.h"
 #include "cmds.h"
 #include "menu.h"
+#include "colors.h"
 
-
-/* Reference count all structs so that items and menus can be shared across
- * menu bars.
- */
-typedef struct
+typedef struct _Menu_Node_Type
 {
    char *name;
    int type;
@@ -43,6 +40,7 @@ typedef struct
 #define MENU_ACTIVE		1
 #define MENU_SELECTION		2
 #define MENU_ITEM_UNAVAILABLE	4
+#define MENU_POPUP_PREPARED	8
 }
 Menu_Node_Type;
 
@@ -53,18 +51,22 @@ typedef struct _Menu_Popup_Type
    unsigned int flags;
 
    /* Private data */
-   /* These 4 must match Menu_Bar_Type */
+   /* These 5 must match Menu_Bar_Type */
    Menu_Node_Type **subnodes;
    unsigned int num_subnodes;
    unsigned int max_subnodes;
    struct _Menu_Popup_Type *parent;
    unsigned int active_node;
 
+   SLang_Name_Type *select_popup_callback;
+   SLang_Name_Type *tweak_popup_callback;
+
    int column;			       /* meaningful when active */
    int row;
    int max_row;
    int max_col;
    unsigned int min_width;
+   int visible_node_offset;
 }
 Menu_Popup_Type;
 
@@ -75,19 +77,21 @@ struct _Menu_Bar_Type
    unsigned int flags;
    
    /* Private data */
-   /* These 4 must match Menu_Popup_Type */
+   /* These 5 must match Menu_Popup_Type */
    Menu_Node_Type **subnodes;
    unsigned int num_subnodes;
    unsigned int max_subnodes;   
    struct _Menu_Popup_Type *parent;   
    unsigned int active_node;
 
-
-   SLang_Name_Type *select_callback;   /* Void select_callback (name) */
    SLang_Name_Type *init_callback;
+   SLang_Name_Type *select_callback;   /* Void select_callback (name) */
 
    int *item_columns;	       /* locations of items on menu bar */
    unsigned int num_refs;
+
+#define DEFAULT_MENU_PREFIX "F10 key ==> "
+   char *prefix_string;
    struct _Menu_Bar_Type *next;
 };
 
@@ -130,8 +134,9 @@ static Menu_Bar_Type *Active_Menu_Bar;
 static Menu_Popup_Type *Active_Popup;
 
 static void free_menu_node (Menu_Node_Type *);
+static int select_menu_cmd (void);
 
-static void free_menu_popup_private_data (Menu_Popup_Type *m)
+static void free_menu_popup_subnodes (Menu_Popup_Type *m)
 {
    Menu_Node_Type **l, **lmax;
 
@@ -144,6 +149,22 @@ static void free_menu_popup_private_data (Menu_Popup_Type *m)
 	l++;
      }
    SLfree ((char *) m->subnodes);
+
+   m->subnodes = NULL;
+   m->num_subnodes = 0;
+   m->max_subnodes = 0;
+}
+
+
+static void free_menu_popup_private_data (Menu_Popup_Type *m)
+{
+   free_menu_popup_subnodes (m);
+#if SLANG_VERSION > 10400
+   SLang_free_function (m->select_popup_callback);
+   SLang_free_function (m->tweak_popup_callback);
+   m->select_popup_callback = NULL;
+   m->tweak_popup_callback = NULL;
+#endif
 }
 
 static void free_menu_bar_private_data (Menu_Bar_Type *m)
@@ -179,6 +200,12 @@ static void free_menu_bar_private_data (Menu_Bar_Type *m)
      }
    SLfree ((char *) m->subnodes);
    SLfree ((char *) m->item_columns);
+   SLang_free_slstring (m->prefix_string);
+#if SLANG_VERSION > 10400
+   SLang_free_function (m->init_callback);
+   SLang_free_function (m->select_callback);
+#endif
+
 }
 
 static void free_keystring_private_data (Menu_Keystring_Fun_Type *m)
@@ -190,6 +217,9 @@ static void free_slangfun_private_data (Menu_SLang_Fun_Type *m)
 {
    if (m->client_data != NULL)
      SLang_free_anytype (m->client_data);
+#if SLANG_VERSION > 10400
+   SLang_free_function (m->slang_fun);
+#endif
 }
 
 static void free_menu_node (Menu_Node_Type *m)
@@ -237,10 +267,9 @@ static Menu_Node_Type *create_menu_node (char *name, int node_type,
 {
    Menu_Node_Type *m;
 
-   m = (Menu_Node_Type *) SLmalloc (sizeof_node);
+   m = (Menu_Node_Type *) jed_malloc0 (sizeof_node);
    if (m == NULL)
      return NULL;
-   memset ((char *)m, 0, sizeof_node);
 
    m->type = node_type;
    
@@ -278,6 +307,7 @@ static Menu_Popup_Type *create_menu_popup (char *name, unsigned int num_items)
 static Menu_Bar_Type *create_menu_bar (char *name, unsigned int num_items)
 {
    Menu_Bar_Type *m;
+   char *prefix;
 
    m = (Menu_Bar_Type *) create_menu_node (name, MENU_NODE_MENUBAR, sizeof (Menu_Bar_Type));
    if (m == NULL)
@@ -287,12 +317,11 @@ static Menu_Bar_Type *create_menu_bar (char *name, unsigned int num_items)
      num_items = 5;
 
    if ((NULL == (m->subnodes = (Menu_Node_Type **)SLmalloc (num_items * sizeof (Menu_Node_Type *))))
-       || (NULL == (m->item_columns = (int *)SLmalloc (num_items * sizeof (int)))))
+       || (NULL == (m->item_columns = (int *)jed_malloc0 (num_items * sizeof (int)))))
      {
 	free_menu_node ((Menu_Node_Type *) m);
 	return NULL;
      }
-   memset ((char *) m->item_columns, 0, num_items * sizeof(int));
 
    m->max_subnodes = num_items;
 
@@ -301,6 +330,20 @@ static Menu_Bar_Type *create_menu_bar (char *name, unsigned int num_items)
    Menu_Bar_Root = m;
 
    m->num_refs = 1;
+   
+   prefix = DEFAULT_MENU_PREFIX;
+   if ((Global_Menu_Bar != NULL)
+       && (Global_Menu_Bar->prefix_string != NULL))
+     prefix = Global_Menu_Bar->prefix_string;
+   
+   prefix = SLang_create_slstring (prefix);
+   if (prefix == NULL)
+     {
+	jed_delete_menu_bar (m);
+	return NULL;
+     }
+   m->prefix_string = prefix;
+
    return m;
 }
 
@@ -409,12 +452,24 @@ static int check_subnode_space (Menu_Popup_Type *m, unsigned int dn)
    return 0;
 }
 
-/* This function assumes that there is enough space */
-static void append_node_to_popup (Menu_Popup_Type *p, Menu_Node_Type *m)
+/* This function assumes that there is enough space to insert one item */
+static void insert_node_to_popup (Menu_Popup_Type *p, Menu_Node_Type *m, 
+				  unsigned int where)
 {
    unsigned int len;
+   unsigned int n;
+   
+   n = p->num_subnodes;
 
-   p->subnodes[p->num_subnodes] = m;
+   if (where > n)
+     where = n;
+
+   while (n > where)
+     {
+	p->subnodes[n] = p->subnodes[n-1];
+	n--;
+     }
+   p->subnodes[where]  = m;
    p->num_subnodes += 1;
 
    /* p could be a menu-bar */
@@ -427,8 +482,14 @@ static void append_node_to_popup (Menu_Popup_Type *p, Menu_Node_Type *m)
      p->min_width = len;
 }
 
+#if 0
+static void append_node_to_popup (Menu_Popup_Type *p, Menu_Node_Type *m)
+{
+   insert_node_to_popup (p, m, p->num_subnodes);
+}
+#endif
    
-static Menu_Popup_Type *append_popup_menu (Menu_Popup_Type *m, char *name)
+static Menu_Popup_Type *insert_popup_menu (Menu_Popup_Type *m, char *name, int where)
 {
    Menu_Popup_Type *p;
 
@@ -444,11 +505,17 @@ static Menu_Popup_Type *append_popup_menu (Menu_Popup_Type *m, char *name)
    
    p->parent = m;
    
-   append_node_to_popup (m, (Menu_Node_Type *)p);
+   insert_node_to_popup (m, (Menu_Node_Type *)p, where);
    return p;
 }
 
-static Menu_Node_Type *append_separator (Menu_Popup_Type *m)
+static Menu_Popup_Type *append_popup_menu (Menu_Popup_Type *m, char *name)
+{
+   return insert_popup_menu (m, name, m->num_subnodes);
+}
+
+
+static Menu_Node_Type *insert_separator (Menu_Popup_Type *m, int where)
 {
    Menu_Node_Type *l;
 
@@ -459,17 +526,23 @@ static Menu_Node_Type *append_separator (Menu_Popup_Type *m)
    if (l == NULL)
      return NULL;
 
-   append_node_to_popup (m, l);
+   insert_node_to_popup (m, l, where);
 
    return l;
 }
 
-static Menu_SLang_Fun_Type *append_slang_fun_item (Menu_Popup_Type *m, char *name, 
-						   SLang_Name_Type *nt, SLang_Any_Type *cd)
+static Menu_Node_Type *append_separator (Menu_Popup_Type *m)
+{
+   return insert_separator (m, m->num_subnodes);
+}
+
+
+static Menu_SLang_Fun_Type *insert_slang_fun_item (Menu_Popup_Type *m, char *name, 
+						   SLang_Name_Type *nt, SLang_Any_Type *cd,
+						   int where)
 {
    Menu_SLang_Fun_Type *l;
 
-
    if (NULL != (l = (Menu_SLang_Fun_Type *)find_subnode (m, name)))
      {
 	if (l->type != MENU_NODE_FUN_SLANG)
@@ -494,11 +567,18 @@ static Menu_SLang_Fun_Type *append_slang_fun_item (Menu_Popup_Type *m, char *nam
    l->slang_fun = nt;
    l->client_data = cd;
 
-   append_node_to_popup (m, (Menu_Node_Type *)l);
+   insert_node_to_popup (m, (Menu_Node_Type *)l, where);
    return l;
 }
 
-static Menu_Keystring_Fun_Type *append_keystring_item (Menu_Popup_Type *m, char *name, char *k)
+static Menu_SLang_Fun_Type *append_slang_fun_item (Menu_Popup_Type *m, char *name,
+						   SLang_Name_Type *nt, SLang_Any_Type *cd)
+{
+   return insert_slang_fun_item (m, name, nt, cd, m->num_subnodes);
+}
+
+static Menu_Keystring_Fun_Type *insert_keystring_item (Menu_Popup_Type *m, char *name, 
+						       char *k, int where)
 {
    Menu_Keystring_Fun_Type *l;
 
@@ -531,10 +611,15 @@ static Menu_Keystring_Fun_Type *append_keystring_item (Menu_Popup_Type *m, char
 	return NULL;
      }
 
-   append_node_to_popup (m, (Menu_Node_Type *)l);
+   insert_node_to_popup (m, (Menu_Node_Type *)l, where);
    return l;
 }
 
+static Menu_Keystring_Fun_Type *append_keystring_item (Menu_Popup_Type *m, char *name, char *k)
+{
+   return insert_keystring_item (m, name, k, m->num_subnodes);
+}
+
 static Menu_Node_Type *get_selected_menu_item (Menu_Popup_Type *p)
 {
    Menu_Node_Type *m;
@@ -833,7 +918,6 @@ static int copy_menu (Menu_Popup_Type *dest, Menu_Node_Type *src)
  *  SLsmg MenuBar interface
  */
 
-#ifndef IBMPC_SYSTEM
 static void simulate_hline (unsigned int n)
 {
    while (n--)
@@ -866,7 +950,22 @@ static void simulate_box (int r, int c, unsigned int dr, unsigned int dc)
 	SLsmg_write_nchars ("|", 1);
      }
 }
-#endif
+
+/* I would prefer to use real up/down arrows but I cannot depend upon their
+ * portability.
+ */
+static void draw_up_arrow (int r, int c)
+{
+   SLsmg_gotorc (r, c);
+   SLsmg_write_string ("(-)");
+}
+
+static void draw_down_arrow (int r, int c)
+{
+   SLsmg_gotorc (r, c);
+   SLsmg_write_string ("(+)");
+}
+
 static void draw_name (char *name, int color0, int color1, unsigned int field_width)
 {
    char *s;
@@ -916,7 +1015,7 @@ static int draw_menu_bar (Menu_Bar_Type *b)
    SLsmg_gotorc (0, 0);
 
    SLsmg_set_color (JMENU_COLOR);
-   SLsmg_write_string ("F10 key ==> ");
+   SLsmg_write_string (b->prefix_string);
 
    m = b->subnodes;
    imax = b->num_subnodes;
@@ -943,7 +1042,7 @@ static int draw_menu_bar (Menu_Bar_Type *b)
 	     if (p->type == MENU_NODE_POPUP)
 	       active = i;
 
-	     draw_name (name, JMENU_SELECTION_COLOR, JMENU_CHAR_COLOR, 0);
+	     draw_name (name, JMENU_SELECTION_COLOR, JMENU_SELECTED_CHAR_COLOR, 0);
 	  }
 	else
 	  draw_name (name, JMENU_COLOR, JMENU_CHAR_COLOR, 0);
@@ -1092,8 +1191,11 @@ static int draw_popup_menu (Menu_Popup_Type *p, int r, int c)
 
    if (r + (int)dr >= Jed_Num_Screen_Rows)
      {
-	if ((int)dr > Jed_Num_Screen_Rows)
-	  r = 0;
+	if ((int)dr >= Jed_Num_Screen_Rows)
+	  {
+	     r = 1;
+	     dr = Jed_Num_Screen_Rows - 1;
+	  }
 	else
 	  r = Jed_Num_Screen_Rows - (int)dr;
      }
@@ -1101,11 +1203,9 @@ static int draw_popup_menu (Menu_Popup_Type *p, int r, int c)
    SLsmg_set_color_in_region (JMENU_SHADOW_COLOR, r + 1, c + 1, dr, dc);
    SLsmg_set_color (JMENU_POPUP_COLOR);
    /* SLsmg_fill_region (r, c, dr, dc, ' '); */
-#ifndef IBMPC_SYSTEM
    if (Jed_Simulate_Graphic_Chars)
      simulate_box (r, c, dr, dc);
    else
-#endif
      SLsmg_draw_box (r, c, dr, dc);
 
    p->row = r;
@@ -1113,15 +1213,38 @@ static int draw_popup_menu (Menu_Popup_Type *p, int r, int c)
    p->max_row = r + dr;
    p->max_col = c + dc;
 
-   l = p->subnodes;
-   lmax = l + p->num_subnodes;
-   
-   c++;
-   r++;
-
    /* Make sure a selection is present */
    (void) get_selected_menu_item (p);
 
+   l = p->subnodes;
+   lmax = l + p->num_subnodes;
+   
+   if (p->num_subnodes + 2 > dr)
+     {
+	unsigned int page;
+	unsigned int nr = dr - 2;
+	unsigned int col;
+	
+	col = c + dc - 4;
+	page = p->active_node / nr;
+	if (page)
+	  {
+	     l += page * nr;
+	     draw_up_arrow (r, col);
+	  }
+	if (lmax > l + nr)
+	  {
+	     lmax = l + nr;
+	     draw_down_arrow (r + dr - 1, col);
+	  }
+	else
+	  l = lmax - nr;
+     }
+   
+   p->visible_node_offset = (int) (l - p->subnodes);
+
+   c++;
+   r++;
    p_active = NULL;
    active_row = r;
    active_col = c;
@@ -1145,6 +1268,7 @@ static int draw_popup_menu (Menu_Popup_Type *p, int r, int c)
 	     active_row = r;
 	     active_col = c + dc;
 	     color0 = JMENU_SELECTION_COLOR;
+	     color1 = JMENU_SELECTED_CHAR_COLOR;
 	  }
 
 	if ((m->flags & MENU_ACTIVE)
@@ -1160,11 +1284,9 @@ static int draw_popup_menu (Menu_Popup_Type *p, int r, int c)
 	  {
 	   case MENU_NODE_SEPARATOR:
 	     SLsmg_write_nchars (" ", 1);
-#ifndef IBMPC_SYSTEM
 	     if (Jed_Simulate_Graphic_Chars)
 	       simulate_hline (dc - 1);
 	     else
-#endif
 	       SLsmg_draw_hline (dc-1);
 	     SLsmg_write_nchars (" ", 1);
 	     break;
@@ -1197,14 +1319,46 @@ static int draw_popup_menu (Menu_Popup_Type *p, int r, int c)
    return 0;
 }
 
+static Menu_Bar_Type *get_active_menubar (void)
+{
+   Menu_Bar_Type *b;
+
+   /* Active_Popup = NULL; */
+   b = Active_Menu_Bar;
+
+   if (b == NULL)
+     {
+	Buffer *buf = CBuf;
+
+	if (IS_MINIBUFFER)
+	  {
+	     if (NULL == (buf = jed_get_mini_action_buffer ()))
+	       return NULL;
+	  }
+
+	if (NULL == (b = buf->menubar))
+	  b = Global_Menu_Bar;
+
+	if (b == NULL)
+	  return NULL;
+	
+     }
+   /* Active_Popup = (Menu_Popup_Type *) b; */
+   return b;
+}
+
 int jed_redraw_menus (void)
 {
    Menu_Popup_Type *p;
    Menu_Bar_Type *b;
    int active;
 
+#if 0
    if (NULL == (b = CBuf->menubar))
      b = Global_Menu_Bar;
+#else
+   b = get_active_menubar ();
+#endif
 
    Active_Popup = (Menu_Popup_Type *) b;
 
@@ -1250,10 +1404,15 @@ static int find_active_popup (void)
      {
 	Active_Popup = p;
 	p = (Menu_Popup_Type *) get_selected_menu_item (p);
-	if ((p == NULL)
-	    || (p->type != MENU_NODE_POPUP)
-	    || (0 == (p->flags & MENU_ACTIVE)))
+
+	if ((p == NULL) || (p->type != MENU_NODE_POPUP))
 	  return 0;
+
+	if (0 == (p->flags & MENU_ACTIVE))
+	  return 0;
+
+	if (0 == (p->flags & MENU_POPUP_PREPARED))
+	  return select_menu_cmd ();
      }
 }
 	
@@ -1267,7 +1426,7 @@ static int next_menubar_cmd (void)
      return -1;
 
    flags = get_active_node_flags (p) & MENU_ACTIVE;
-
+   
    /* (void) unselect_active_node ((Menu_Popup_Type *)Active_Menu_Bar); */
    (void) select_next_active_node (p, p->active_node, flags);
    (void) find_active_popup ();
@@ -1283,7 +1442,6 @@ static int prev_menubar_cmd (void)
      return -1;
 
    flags = get_active_node_flags (p) & MENU_ACTIVE;
-
    /* (void) unselect_active_node ((Menu_Popup_Type *)Active_Menu_Bar); */
    (void) select_prev_active_node (p, p->active_node, flags);
    (void) find_active_popup ();
@@ -1296,9 +1454,11 @@ int jed_exit_menu_bar (void)
      {
 	unselect_active_node (Active_Popup);
 	Active_Popup->active_node = 0;
+	Active_Popup->flags &= ~MENU_POPUP_PREPARED;
 	Active_Popup = Active_Popup->parent;
      }
 
+   Active_Menu_Bar = NULL;
    Jed_Menus_Active = 0;
    return 1;
 }
@@ -1355,8 +1515,83 @@ static int execute_keystring (char *s)
    return 1;
 }
 
-	
+static char *get_full_popup_name (Menu_Popup_Type *p)
+{
+   Menu_Popup_Type *parent;
+   unsigned int len;
+   char *name;
+   char *s;
+
+   len = strlen (p->name);
+   parent = p->parent;
+   while (parent != NULL)
+     {
+	len += 1 + strlen (parent->name);
+	parent = parent->parent;
+     }
    
+   name = SLmalloc (len + 1);
+   if (name == NULL)
+     return NULL;
+   
+   s = (name + len) - strlen (p->name);
+   strcpy (s, p->name);
+   parent = p->parent;
+   while (parent != NULL)
+     {
+	s--;
+	*s = '.';
+	len = strlen (parent->name);
+	s -= len;
+	strncpy (s, parent->name, len);
+	parent = parent->parent;
+     }
+   
+   return name;
+}
+
+static int run_popup_callback (Menu_Popup_Type *p, SLang_Name_Type *cb)
+{
+   char *name;
+
+   if (NULL == (name = get_full_popup_name (p)))
+     return -1;
+
+   if (-1 == SLang_push_string (name))
+     {
+	SLfree (name);
+	return -1;
+     }
+   SLfree (name);
+
+   if (0 != SLexecute_function (cb))
+     return -1;
+
+   return 0;
+}
+
+   
+static int prepare_popup (Menu_Popup_Type *p)
+{
+   SLang_Name_Type *cb;
+
+   if (NULL != (cb = p->select_popup_callback))
+     {
+	free_menu_popup_subnodes (p);
+	if (-1 == run_popup_callback (p, cb))
+	  return -1;
+     }
+   
+   if (NULL != (cb = p->tweak_popup_callback))
+     {
+	if (-1 == run_popup_callback (p, cb))
+	  return -1;
+     }
+   
+   p->flags |= MENU_POPUP_PREPARED;
+   return 0;
+}
+
 
 static int select_menu_cmd (void)
 {
@@ -1373,8 +1608,16 @@ static int select_menu_cmd (void)
    switch (m->type)
      {
       case MENU_NODE_POPUP:
+#if 0
 	m->flags |= MENU_ACTIVE;
 	p = (Menu_Popup_Type *) m;
+#else
+	/* Do it this way to avoid DEC C ansi-aliasing warning. */
+	p = (Menu_Popup_Type *) m;
+	p->flags |= MENU_ACTIVE;
+#endif
+	if (-1 == prepare_popup (p))
+	  return -1;
 	Active_Popup = p;
 	select_next_active_node (p, p->num_subnodes, 0);
 	return 1;
@@ -1415,7 +1658,7 @@ static int back_menu_cmd (void)
        || (Active_Popup->parent == (Menu_Popup_Type *) Active_Menu_Bar))
      return jed_exit_menu_bar ();
 
-   Active_Popup->flags &= ~MENU_ACTIVE;
+   Active_Popup->flags &= ~(MENU_ACTIVE|MENU_POPUP_PREPARED);
    Active_Popup = Active_Popup->parent;
    return 1;
 }
@@ -1494,7 +1737,9 @@ static SLKeymap_Function_Type Menu_Keymap_Functions [] =
 };
 
 #ifndef IBMPC_SYSTEM
+# ifdef HAS_MOUSE
 static int xterm_mouse_cmd (void);
+# endif
 #endif
 
 static int init_menu_keymap (void)
@@ -1548,7 +1793,9 @@ static int init_menu_keymap (void)
    SLkm_define_key ("^(kP)", (FVOID_STAR) pgup_menu_cmd, Menu_Keymap);
    SLkm_define_key ("^(kN)", (FVOID_STAR) pgdn_menu_cmd, Menu_Keymap);
 # endif
+# ifdef HAS_MOUSE
    SLkm_define_key ("\033[M", (FVOID_STAR) xterm_mouse_cmd, Menu_Keymap);
+# endif
 #else				       /* IBMPC_SYSTEM */
 #include "doskeys.h"
    
@@ -1582,6 +1829,13 @@ SLang_Key_Type *jed_menu_do_key (void)
      return SLang_do_key (Menu_Keymap, jed_getkey);
 
    ch = my_getkey ();
+
+   if (SLKeyBoard_Quit || SLang_Error)
+     {
+	jed_exit_menu_bar ();
+	return NULL;
+     }
+
    if (ch == 27)		       /* ESC */
      {
 	int tsecs = 2;
@@ -1593,22 +1847,6 @@ SLang_Key_Type *jed_menu_do_key (void)
    return SLang_do_key (Menu_Keymap, jed_getkey);
 }
 
-static Menu_Bar_Type *get_active_menubar (void)
-{
-   Active_Popup = NULL;
-   if (Active_Menu_Bar == NULL)
-     {
-	if (NULL == (Active_Menu_Bar = CBuf->menubar))
-	  Active_Menu_Bar = Global_Menu_Bar;
-
-	if (Active_Menu_Bar == NULL)
-	  return NULL;
-     }
-
-   Active_Popup = (Menu_Popup_Type *) Active_Menu_Bar;
-   return Active_Menu_Bar;
-}
-
 static int exec_menubar_callback (Menu_Bar_Type *b, SLang_Name_Type **ntp)
 {
    SLang_Name_Type *nt;
@@ -1617,13 +1855,21 @@ static int exec_menubar_callback (Menu_Bar_Type *b, SLang_Name_Type **ntp)
    if (nt == NULL)
      return 0;
 
-   if (-1 == SLang_push_string (b->name))
+   if (SLang_Error)
      return -1;
 
+   if (-1 == SLang_push_string (b->name))
+     return -1;
+   
    if (-1 == SLexecute_function (nt))
      {
 	if (Active_Menu_Bar == b)      /* disable callback */
-	  *ntp = NULL;
+	  {
+#if SLANG_VERSION > 10400
+	     SLang_free_function (nt);
+#endif
+	     *ntp = NULL;
+	  }
 	return -1;
      }
 
@@ -1634,11 +1880,16 @@ static int exec_menubar_callback (Menu_Bar_Type *b, SLang_Name_Type **ntp)
 void jed_notify_menu_buffer_changed (void)
 {
    Menu_Bar_Type *b;
-   
+   int err;
+
    if (NULL == (b = get_active_menubar ()))
      return;
 
+   err = SLang_Error;
+   SLang_Error = 0;
    (void) exec_menubar_callback (b, &b->init_callback);
+   if (SLang_Error == 0)
+     SLang_Error = err;
 }
 
 
@@ -1647,8 +1898,19 @@ int jed_select_menu_bar (void)
    Menu_Popup_Type *p;
    Menu_Node_Type **l, **lmax;
 
-   if (NULL == (p = (Menu_Popup_Type *) get_active_menubar ()))
+   if (IS_MINIBUFFER)
+     {
+	/* For now do not allow access to menus from the minibuffer
+	 * since the minibuffer is not recursive
+	 */
+	return -1;
+     }
+   
+   if (NULL == (Active_Menu_Bar = get_active_menubar ()))
      return -1;
+   
+   p = (Menu_Popup_Type *) Active_Menu_Bar;
+   Active_Popup = p;
 
    l = p->subnodes;
    lmax = l + p->num_subnodes;
@@ -1671,7 +1933,10 @@ int jed_select_menu_bar (void)
    p->active_node = 0;
    
    if (-1 == exec_menubar_callback (Active_Menu_Bar, &Active_Menu_Bar->select_callback))
-     return -1;
+     {
+	jed_exit_menu_bar ();
+	return -1;
+     }
    
    (void) get_selected_menu_item (p);
    touch_screen ();
@@ -1766,16 +2031,19 @@ static Menu_Popup_Type *find_menu_popup (char *name)
    return m;
 }
 
-
-   
-static void append_popup_menu_cmd (char *dest, char *menu)
+/* If name is NULL or non-existent, then insert at end */
+static unsigned int find_where_to_insert (Menu_Popup_Type *p, char *name)
 {
-   Menu_Popup_Type *m;
+   unsigned int i, num;
 
-   m = find_menu_popup (dest);
+   num = p->num_subnodes;
 
-   if (m != NULL)
-     (void) append_popup_menu (m, menu);
+   if (name != NULL) for (i = 0; i < num; i++)
+     {
+	if (0 == strcmp (name, p->subnodes[i]->name))
+	  return i;
+     }
+   return num;
 }
 
 static void create_menu_bar_cmd (char *name)
@@ -1809,6 +2077,70 @@ static void set_buffer_menu_bar_cmd (char *name)
    b->num_refs += 1;
 }
 
+static int pop_where_to_insert (Menu_Popup_Type *p, unsigned int *where)
+{
+   if (SLang_peek_at_stack () == SLANG_STRING_TYPE)
+     {
+	char *s;
+	
+	if (-1 == SLang_pop_slstring (&s))
+	  return -1;
+	*where = find_where_to_insert (p, s);
+	SLang_free_slstring (s);
+	return 0;
+     }
+   
+   if (-1 == SLang_pop_uinteger (where))
+     return -1;
+   
+   return 0;
+}
+
+static void insert_popup_menu_cmd (char *dest, char *menu)
+{
+   Menu_Popup_Type *m;
+   unsigned int where;
+
+   m = find_menu_popup (dest);
+
+   if (m == NULL)
+     return;
+
+   if (-1 == pop_where_to_insert (m, &where))
+     return;
+
+   insert_popup_menu (m, menu, where);
+}
+
+   
+static void append_popup_menu_cmd (char *dest, char *menu)
+{
+   Menu_Popup_Type *m;
+
+   m = find_menu_popup (dest);
+
+   if (m == NULL)
+     return;
+
+   insert_popup_menu (m, menu, m->num_subnodes);
+}
+
+static void insert_separator_cmd (char *name)
+{
+   Menu_Popup_Type *m;
+   unsigned int where;
+
+   m = find_menu_popup (name);
+
+   if (m == NULL)
+     return;
+
+   if (-1 == pop_where_to_insert (m, &where))
+     return;
+
+   insert_separator (m, where);
+}
+
 static void append_separator_cmd (char *name)
 {
    Menu_Popup_Type *m;
@@ -1816,10 +2148,10 @@ static void append_separator_cmd (char *name)
    m = find_menu_popup (name);
 
    if (m != NULL)
-     (void) append_separator (m);
+     insert_separator (m, m->num_subnodes);
 }
 
-static void append_menu_item_cmd (void)
+static void insert_menu_item_cmd_internal (int is_fun, int do_append)
 {
    Menu_Popup_Type *m;
    SLang_Name_Type *nt = NULL;
@@ -1827,8 +2159,9 @@ static void append_menu_item_cmd (void)
    char *str = NULL;
    char *menu = NULL;
    char *name = NULL;
+   unsigned int where;
 
-   if (SLang_Num_Function_Args == 4)
+   if (is_fun)
      {
 	if (-1 == SLang_pop_anytype (&client_data))
 	  return;
@@ -1847,14 +2180,22 @@ static void append_menu_item_cmd (void)
    
    if (NULL == (m = find_menu_popup (menu)))
      goto free_and_return;
+   
+   if (do_append)
+     where = m->num_subnodes;
+   else if (-1 == pop_where_to_insert (m, &where))
+     goto free_and_return;
 
    if (nt != NULL)
      {
-	if (NULL != append_slang_fun_item (m, name, nt, client_data))
-	  client_data = NULL;
+	if (NULL != insert_slang_fun_item (m, name, nt, client_data, where))
+	  {
+	     client_data = NULL;
+	     nt = NULL;
+	  }
      }
    else
-     (void) append_keystring_item (m, name, str);
+     (void) insert_keystring_item (m, name, str, where);
    /* drop */
 
    free_and_return:
@@ -1864,6 +2205,19 @@ static void append_menu_item_cmd (void)
    SLang_free_slstring (str);
    SLang_free_slstring (menu);
    SLang_free_slstring (name);
+#if SLANG_VERSION > 10400
+   SLang_free_function (nt);
+#endif
+}
+
+static void insert_menu_item_cmd (void)
+{
+   insert_menu_item_cmd_internal ((SLang_Num_Function_Args == 5), 0);
+}
+
+static void append_menu_item_cmd (void)
+{
+   insert_menu_item_cmd_internal ((SLang_Num_Function_Args == 4), 1);
 }
 
 static void menu_delete_item_cmd (char *menu)
@@ -1904,11 +2258,12 @@ static void set_object_available_cmd (char *name, int *flag)
      m->flags |= MENU_ITEM_UNAVAILABLE;
 }
 
-static int pop_menubar_callback (Menu_Bar_Type **bp, SLang_Name_Type **ntp)
+static int pop_popup_callback (Menu_Popup_Type **pp, int type,
+			       SLang_Name_Type **ntp)
 {
    SLang_Name_Type *nt;
-   char *menubar;
-   Menu_Bar_Type *b;
+   char *popup;
+   Menu_Popup_Type *p;
 
    if (SLang_peek_at_stack () == SLANG_NULL_TYPE)
      {
@@ -1918,27 +2273,51 @@ static int pop_menubar_callback (Menu_Bar_Type **bp, SLang_Name_Type **ntp)
    else if (NULL == (nt = SLang_pop_function ()))
      return -1;
 
-   if (-1 == SLang_pop_slstring (&menubar))
+   if (-1 == SLang_pop_slstring (&popup))
      {
+#if SLANG_VERSION > 10400
+	SLang_free_function (nt);
+#endif
+	return -1;
+     }
+
+   p = find_menu_popup (popup);
+   if (p == NULL)
+     {
+#if SLANG_VERSION > 10400
+	SLang_free_function (nt);
+#endif
+	SLang_free_slstring (popup);
 	return -1;
      }
    
-   b = menu_find_menu_bar (menubar, 1);
-   SLang_free_slstring (menubar);
-
-   if (b == NULL)
-     return -1;
+   if (type && (p->type != type))
+     {
+	SLang_verror (SL_INVALID_PARM, "%s does not specify the proper menu type", popup);
+#if SLANG_VERSION > 10400
+	SLang_free_function (nt);
+#endif
+	SLang_free_slstring (popup);
+	return -1;
+     }
+   
+   SLang_free_slstring (popup);
 
    *ntp = nt;
-   *bp = b;
+   *pp = p;
    return 0;
 }
 
+static int pop_menubar_callback (Menu_Bar_Type **bp, SLang_Name_Type **nt)
+{
+   return pop_popup_callback ((Menu_Popup_Type **)bp, MENU_NODE_MENUBAR, nt);
+}
+
 static void set_select_menubar_callback (void)
 {
    Menu_Bar_Type *b;
    SLang_Name_Type *nt;
-   
+
    if (-1 == pop_menubar_callback (&b, &nt))
      return;
 
@@ -1956,6 +2335,29 @@ static void set_init_menubar_callback (void)
    b->init_callback = nt;
 }
 
+
+static void set_select_popup_callback (void)
+{
+   Menu_Popup_Type *p;
+   SLang_Name_Type *nt;
+
+   if (-1 == pop_popup_callback (&p, MENU_NODE_POPUP, &nt))
+     return;
+
+   p->select_popup_callback = nt;
+}
+
+static void set_tweak_popup_callback (void)
+{
+   Menu_Popup_Type *p;
+   SLang_Name_Type *nt;
+
+   if (-1 == pop_popup_callback (&p, MENU_NODE_POPUP, &nt))
+     return;
+
+   p->tweak_popup_callback = nt;
+}
+
 static void copy_menu_cmd (char *destname, char *srcname)
 {
    Menu_Popup_Type *dest;
@@ -1972,20 +2374,84 @@ static void copy_menu_cmd (char *destname, char *srcname)
    (void) copy_menu (dest, src);
 }
 
+static void set_menu_bar_prefix_string (char *menubar, char *s)
+{
+   Menu_Bar_Type *b;
+   
+   if (NULL == (b = menu_find_menu_bar (menubar, 1)))
+     return;
+
+   s = SLang_create_slstring (s);
+   if (s == NULL)
+     return;
+   
+   SLang_free_slstring (b->prefix_string);
+   b->prefix_string = s;
+}
+
+   
+static void popup_menu_cmd (char *menu)
+{
+   char *popup;
+
+   (void) jed_exit_menu_bar ();
+   if (-1 == jed_select_menu_bar ())
+     return;
+
+   if (NULL == (menu = SLmake_string (menu)))
+     return;
+
+   /* for now ignore the menubar name and use default */
+   popup = strchr (menu, '.');
+   if (popup != NULL)
+     popup++;
+
+   while ((popup != NULL) && (Active_Popup != NULL))
+     {
+	Menu_Node_Type *m;
+	char *next_popup;
+	
+	next_popup = strchr (popup, '.');
+	if (next_popup != NULL)
+	  *next_popup++ = 0;
+
+	if (NULL == (m = find_subnode (Active_Popup, popup)))
+	  {
+	     SLang_verror (SL_INVALID_PARM, 
+			   "Unable to find a popup menu called %s", menu);
+	     break;
+	  }
+	set_node_selection (Active_Popup, m);
+	if (-1 == select_menu_cmd ())
+	  break;
+	
+	popup = next_popup;
+     }
+   
+   SLfree (menu);
+}
+
+
 static SLang_Intrin_Fun_Type Menu_Table[] =
 {
    MAKE_INTRINSIC_S("menu_create_menu_bar", create_menu_bar_cmd, VOID_TYPE),
    MAKE_INTRINSIC_SS("menu_append_popup", append_popup_menu_cmd, VOID_TYPE),
+   MAKE_INTRINSIC_SS("menu_insert_popup", insert_popup_menu_cmd, VOID_TYPE),
    MAKE_INTRINSIC_S("menu_use_menu_bar", set_buffer_menu_bar_cmd, VOID_TYPE),
    MAKE_INTRINSIC_S("menu_append_separator", append_separator_cmd, VOID_TYPE),
+   MAKE_INTRINSIC_S("menu_insert_separator", insert_separator_cmd, VOID_TYPE),
    MAKE_INTRINSIC_0("menu_append_item", append_menu_item_cmd, VOID_TYPE),
+   MAKE_INTRINSIC_0("menu_insert_item", insert_menu_item_cmd, VOID_TYPE),
    MAKE_INTRINSIC_S("menu_delete_item", menu_delete_item_cmd, VOID_TYPE),
    MAKE_INTRINSIC_S("menu_delete_items", menu_delete_items_cmd, VOID_TYPE),
    MAKE_INTRINSIC_SI("menu_set_object_available", set_object_available_cmd, VOID_TYPE),
    MAKE_INTRINSIC_0("menu_set_select_menubar_callback", set_select_menubar_callback, VOID_TYPE),
    MAKE_INTRINSIC_0("menu_set_init_menubar_callback", set_init_menubar_callback, VOID_TYPE),
+   MAKE_INTRINSIC_0("menu_set_select_popup_callback", set_select_popup_callback, VOID_TYPE),
+   MAKE_INTRINSIC_0("menu_set_tweak_popup_callback", set_tweak_popup_callback, VOID_TYPE),
    MAKE_INTRINSIC_SS("menu_copy_menu", copy_menu_cmd, VOID_TYPE),
-
+   MAKE_INTRINSIC_SS("menu_set_menu_bar_prefix", set_menu_bar_prefix_string, VOID_TYPE),
+   MAKE_INTRINSIC_S("menu_select_menu", popup_menu_cmd, VOID_TYPE),
    MAKE_INTRINSIC(NULL,NULL,0,0)
 };
 
@@ -2003,6 +2469,7 @@ int jed_init_menus (void)
    return 0;
 }
 
+#ifdef HAS_MOUSE
 static int select_menu_via_rc (int type, int r, int c)
 {
    Menu_Popup_Type *p;
@@ -2031,7 +2498,7 @@ static int select_menu_via_rc (int type, int r, int c)
 	  {
 	     if (type != JMOUSE_DRAG)
 	       return -1;
-	     
+
 	     if (Active_Popup->type != MENU_NODE_MENUBAR)
 	       return 0;
 	  }
@@ -2070,6 +2537,7 @@ static int select_menu_via_rc (int type, int r, int c)
    if (p == Active_Popup)
      {
 	r -= (p->row + 1);
+	r += p->visible_node_offset;
 	if ((r >= (int)p->num_subnodes) || (r < 0))
 	  return 0;
 	
@@ -2090,7 +2558,6 @@ static int select_menu_via_rc (int type, int r, int c)
    return 0;
 }
 
-
 int jed_menu_handle_mouse (unsigned int type, 
 			   int x, int y, int button, int shift)
 {
@@ -2099,8 +2566,9 @@ int jed_menu_handle_mouse (unsigned int type,
    if ((type != JMOUSE_UP) && (type != JMOUSE_DRAG))
      return -1;
 
-   if (Jed_Menus_Active == 0)
-     jed_select_menu_bar ();
+   if ((Jed_Menus_Active == 0)
+       && (-1 == jed_select_menu_bar ()))
+     return -1;
 
    if (-1 == select_menu_via_rc (type, y-1, x-1))
      {
@@ -2130,6 +2598,6 @@ static int xterm_mouse_cmd (void)
 }
 
 
-#endif
-
+#endif 				       /* !IBMPC_SYSTEM */
+#endif				       /* HAS_MOUSE */
 #endif				       /* JED_HAS_TTY_MENUS */
