@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -43,8 +43,6 @@
 #include "jprocess.h"
 #endif
 
-int Abort_Char = 7;		       /* scan code for G (control) */
-
 #include "dos_os2.c"
 
 extern char *get_cwd (void);
@@ -155,7 +153,7 @@ static void os2_kbdhandler_thread (void)
 	  {
 	     /* found a char process it */
 
-	     if (keyInfo.chChar == Abort_Char)
+	     if (keyInfo.chChar == Jed_Abort_Char)
 	       {
 		  if (Ignore_User_Abort == 0) SLang_Error = USER_BREAK;
 		  SLKeyBoard_Quit = 1;
@@ -366,7 +364,7 @@ int sys_chmod(char *file, int what, int *mode, short *uid, short *gid)
 # ifdef _MSC_VER
    /* MSC stat() is broken on directory names ending with a slash */
    char path[_MAX_PATH];
-   strcpy(path, file);
+   safe_strcpy(path, file, sizeof (path));
    deslash(file = path);
 # endif
 
@@ -402,7 +400,7 @@ int sys_chmod(char *file, int what, int *mode, short *dum1, short *dum2)
    (void) dum1; (void) dum2;
 
  /* DosQueryPathInfo won't work with directory names ending with a slash */
-   strcpy(path, file);
+   safe_strcpy(path, file, sizeof(path));
    deslash(file = path);
    Result = DosQueryPathInfo (path, FIL_STANDARD, &fileInfo, sizeof (fileInfo));
    if (Result == ERROR_FILE_NOT_FOUND) return 0;
@@ -485,8 +483,9 @@ int sys_findfirst(char *theFile)
     * Makefile.in between the others, so check for wrong values and reject
     * them 
     */
-   if ((strncmp(theFile, Found_File, Found_FLen) != 0) &&
-       (!sys_findnext(theFile)))
+   if (Jed_Filename_Case_Sensitive
+       && (0 != strncmp (file, Found_File, Found_FLen))
+       && !sys_findnext (file))
      return 0;
 
    if (FindBuffer.attrFile & FILE_DIRECTORY) fixup_dir (theFile);
@@ -518,8 +517,10 @@ int sys_findnext(char *file)
 	     * case-insensitive so if one searches for makefi*, it will return
 	     * Makefile.in between the others, so check for wrong values and
 	     * reject them */
-	     if (strncmp(file, Found_File, Found_FLen) != 0)
-	      continue;
+	     if (Jed_Filename_Case_Sensitive
+		 && (0 != strncmp (file, Found_File, Found_FLen)))
+	       continue;
+
 	     if (FindBuffer.attrFile & FILE_DIRECTORY) fixup_dir (file);
 	     return 1;
 	  }
