@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -31,6 +31,9 @@
 #include "file.h"
 #include "hooks.h"
 #include "menu.h"
+#include "version.h"
+#include "indent.h"
+#include "colors.h"
 
 #if JED_HAS_SUBPROCESSES
 # include "jprocess.h"
@@ -38,16 +41,10 @@
 
 #include "sig.h"
 
-#define JED_VERSION "B0.99.9"
-int Jed_Version_Number = 9909;
-char *Jed_Version_String = JED_VERSION;
-
 volatile int Jed_Resize_Pending;
 char *MiniBuf_Get_Response_String;
 
-#ifndef IBMPC_SYSTEM
 int Jed_Simulate_Graphic_Chars;
-#endif
 
 typedef struct Screen_Type
 {
@@ -68,7 +65,7 @@ int Cursor_Motion;    /* indicates cursor movement only -1 ^ v +1 < > */
 int Jed_Dollar = '$';
 int User_Prefers_Line_Numbers = 0;
 int Mode_Has_Syntax_Highlight;
-int Wants_Syntax_Highlight;	       /* if non-zero, highlight the syntax.
+int Wants_Syntax_Highlight = 1;	       /* if non-zero, highlight the syntax.
 					*/
 int Wants_Attributes = 1;
 int Wants_HScroll = 20;		       /* controls automatic horizontal
@@ -95,6 +92,7 @@ static int HScroll;		       /* amount to scroll line by */
 static int Absolute_Column;
 
 static int Point_Cursor_Flag = 1;      /* if non-zero, point cursor */
+int Jed_Display_Initialized;
 
 static Line Eob_Line =
 {
@@ -102,6 +100,9 @@ static Line Eob_Line =
 #ifdef KEEP_SPACE_INFO
      ,5
 #endif
+#if JED_HAS_LINE_ATTRIBUTES
+     , JED_LINE_IS_READONLY
+#endif
 };
 
 static unsigned char Char_Width[256];
@@ -111,7 +112,7 @@ static int Display_Eight_Bit = 0x7FFF;
      if (SLsmg_Display_Eight_Bit != Display_Eight_Bit) fix_char_width ()
 
 
-static void display_line (Line *line, int row)
+static void display_line (Line *line, int sy, int sx)
 {
    unsigned int len;
    int hscroll_col;
@@ -122,15 +123,13 @@ static void display_line (Line *line, int row)
 #endif
    
    SLsmg_Tab_Width = Buffer_Local.tab;
-   is_mini = (row == Jed_Num_Screen_Rows);
+   is_mini = (sy + 1 == Jed_Num_Screen_Rows);
    
-   row--;
-   
-   SLsmg_gotorc (row, 0);
+   SLsmg_gotorc (sy, sx);
    SLsmg_set_color (0);
    
-   s = JScreen + row;
-   
+   s = JScreen + sy;
+
    s->line = line;
    s->is_modified = 0;
    
@@ -140,15 +139,15 @@ static void display_line (Line *line, int row)
 	return;
      }
    
-   hscroll_col = JWindow->column - 1;
+   hscroll_col = JWindow->hscroll_column - 1;
    
    if ((line == HScroll_Line)
        && Wants_HScroll && HScroll)
      hscroll_col += HScroll;
    
-   if (hscroll_col)
+   if (hscroll_col + sx)
      {
-	int tmp = hscroll_col;
+	int tmp = hscroll_col + sx;
 	SLsmg_set_screen_start (NULL, &tmp);
      }
    
@@ -204,22 +203,24 @@ static void display_line (Line *line, int row)
 #endif
    SLsmg_erase_eol ();
    
-   if (hscroll_col + Jed_Num_Screen_Cols <= SLsmg_get_column ())
-     {
-	char dollar = (char) Jed_Dollar;
-	SLsmg_gotorc (row, hscroll_col + Jed_Num_Screen_Cols - 1);
-	SLsmg_set_color (JDOLLAR_COLOR);
-	SLsmg_write_nchars (&dollar, 1);
-     }
-   if (hscroll_col)
+   if (Jed_Dollar)
      {
 	char dollar = (char) Jed_Dollar;
 	
-	SLsmg_gotorc (row, hscroll_col);
-	SLsmg_set_color (JDOLLAR_COLOR);
-	SLsmg_write_nchars (&dollar, 1);
+	if (sx + hscroll_col + Jed_Num_Screen_Cols <= SLsmg_get_column ())
+	  {
+	     SLsmg_gotorc (sy, hscroll_col + Jed_Num_Screen_Cols - 1);
+	     SLsmg_set_color (JDOLLAR_COLOR);
+	     SLsmg_write_nchars (&dollar, 1);
+	  }
+	if (hscroll_col)
+	  {
+	     SLsmg_gotorc (sy, sx + hscroll_col);
+	     SLsmg_set_color (JDOLLAR_COLOR);
+	     SLsmg_write_nchars (&dollar, 1);
+	  }
      }
-   
+
    if ((s->hi0 != NULL) && Wants_Attributes)
      {
 	int c;
@@ -233,13 +234,13 @@ static void display_line (Line *line, int row)
 	     c = jed_compute_effective_length (line->data, s->hi0);
 	     if (is_mini)
 	       c += Mini_Info.effective_prompt_len;
-	     SLsmg_gotorc (row, c);
+	     SLsmg_gotorc (sy, c + sx);
 	     SLsmg_set_color (JREGION_COLOR);
 	     SLsmg_write_nchars ((char *)s->hi0, len);
 	  }
      }
    
-   if (hscroll_col)
+   if (hscroll_col + sx)
      SLsmg_set_screen_start (NULL, NULL);
    
    SLsmg_set_color (0);
@@ -287,7 +288,7 @@ Line *jed_find_non_hidden_line (Line *l)
 
 #endif
 
-static Line *find_top_to_recenter (Line *cline)
+Line *jed_find_top_to_recenter (Line *cline)
 {
    int n;
    Line *prev, *last_prev;
@@ -339,10 +340,10 @@ Line *find_top (void)
    /* Note: top_window_line might be a bogus pointer.  This means that I cannot
     * access it unless it really corresponds to a pointer in the buffer.
     */
-   top_window_line = JScreen [JWindow->top - 1].line;
+   top_window_line = JScreen [JWindow->sy].line;
    
    if (top_window_line == NULL)
-     return find_top_to_recenter (cline);
+     return jed_find_top_to_recenter (cline);
    
    /* Chances are that the current line is visible in the window.  This means
     * that the top window line should be above it.
@@ -376,7 +377,7 @@ Line *find_top (void)
    
    if ((*tt_Term_Cannot_Scroll)
        && (*tt_Term_Cannot_Scroll != -1))
-     return find_top_to_recenter (cline);
+     return jed_find_top_to_recenter (cline);
    
    next = cline->next;
 #if JED_HAS_LINE_ATTRIBUTES
@@ -394,11 +395,11 @@ Line *find_top (void)
      prev = prev->prev;
 #endif
    
-   top_window_line = JScreen [nrows + JWindow->top - 2].line;
+   top_window_line = JScreen [nrows + JWindow->sy - 1].line;
    
    if ((prev == NULL)
        || (prev != top_window_line))
-     return find_top_to_recenter (cline);
+     return jed_find_top_to_recenter (cline);
    
    /* It looks like cline is below window by one line.  See what line should
     * be at top to scroll it into view.
@@ -423,7 +424,7 @@ Line *find_top (void)
    if (prev != NULL)
      return prev;
    
-   return find_top_to_recenter (cline);
+   return jed_find_top_to_recenter (cline);
 }
 
 /* This function has to be consistent with SLsmg. */
@@ -436,18 +437,16 @@ static void fix_char_width(void)
    
    /* Control Characters */
    for (i = 0; i < 32; i++)
-     {
-	Char_Width[i] = 2;
-	Char_Width[i + 128] = 3;
-     }
+     Char_Width[i] = 2;
    
    for (i = 32; i < 127; i++)
-     {
-	Char_Width[i] = 1;
-	Char_Width[i + 128] = 1;
-     }
+     Char_Width[i] = 1;
+
    Char_Width[127] = 2;
    
+   for (i = 128; i < SLsmg_Display_Eight_Bit; i++)
+     Char_Width[i] = 3;
+
    for (i = SLsmg_Display_Eight_Bit; i < 256; i++)
      Char_Width[i] = 1;
    
@@ -543,9 +542,9 @@ void point_cursor (int c)
 {
    int r, row;
    Line *tthis, *cline;
-   
+
    if (JWindow->trashed) return;
-   
+
    cline = CLine;
 #if JED_HAS_LINE_ATTRIBUTES
    if (cline->flags & JED_LINE_HIDDEN)
@@ -556,7 +555,7 @@ void point_cursor (int c)
      }
 #endif
    
-   r = JWindow->top;
+   r = JWindow->sy + 1;
    Point_Cursor_Flag = 0;
    for (row = r; row < r + JWindow->rows; row++)
      {
@@ -580,11 +579,13 @@ void point_cursor (int c)
    
    if (c == 0)
      c = calculate_column ();
-   c -= (JWindow->column - 1);
+   c -= (JWindow->hscroll_column - 1);
    
    if (cline == HScroll_Line) c -= HScroll;
    if (c < 1) c = 1; else if (c > JWindow->width) c = JWindow->width;
    
+   c += JWindow->sx;
+
    SLsmg_gotorc (r - 1, c - 1);
    Screen_Row = r;
    Screen_Col = c;
@@ -653,7 +654,7 @@ static void finish_status(int col_flag)
    char *v, ch;
    Line *l;
    int top, rows, narrows;
-   char buf [128];
+   char buf [256];
    char *str;
    
    v = CBuf->status_line;
@@ -674,6 +675,11 @@ static void finish_status(int col_flag)
 	
 	switch (ch)
 	  {
+	   case 'F':
+	     SLsmg_write_string (CBuf->dir);
+	     str = CBuf->file;
+	     break;
+
 	   case 'a':
 	     if (CBuf->flags & ABBREV_MODE) str = " abbrev"; else str = NULL;
 	     break;
@@ -691,13 +697,17 @@ static void finish_status(int col_flag)
 	   case 'o':
 	     if (CBuf->flags & OVERWRITE_MODE) str = " Ovwrt"; else str = NULL;
 	     break;
+	   case 'O':
+	     if (CBuf->flags & OVERWRITE_MODE) str = " ovr"; else str = " ins";
+	     break;
+
 	   case 'b': str = CBuf->name; break;
 	     
 	   case 'p':
 	     str = buf;
 	     if (0 == User_Prefers_Line_Numbers)
 	       {
-		  top = JWindow->top - 1;	
+		  top = JWindow->sy;	
 		  rows = JWindow->rows - 1;
 		  l = JScreen[top + rows].line;
 		  if (l == CBuf->end) l = NULL;
@@ -725,7 +735,17 @@ static void finish_status(int col_flag)
 	       }
 	     break;
 	     
-	   case 'v': str = JED_VERSION; break;
+	   case 'l':
+	     sprintf(buf, "%d", LineNum);
+	     str=buf;
+	     break;
+        
+	   case 'L':
+	     sprintf(buf, "%d", Max_LineNum);
+	     str=buf;
+	     break;
+
+	   case 'v': str = Jed_Version_String; break;
 	   case 'm': str = CBuf->mode_string; break;
 	   case 't': str = status_get_time(); break;
 	   case 'c':
@@ -760,13 +780,13 @@ static int update_status_line (int col_flag)
    char buf[32], *b;
    int num;
    
-   if (JWindow->top == Jed_Num_Screen_Rows) return 0;   /* minibuffer ? */
+   if (JWindow->sy + 1 == Jed_Num_Screen_Rows) return 0;   /* minibuffer ? */
    
-   SLsmg_gotorc (JWindow->rows + JWindow->top - 1, 0);
+   SLsmg_gotorc (JWindow->rows + JWindow->sy, 0);
    SLsmg_set_color (JSTATUS_COLOR);
    
    b = buf;
-   if (JWindow->column != 1) *b++ = '<'; else *b++ = '-';
+   if (JWindow->hscroll_column != 1) *b++ = '<'; else *b++ = '-';
    
    if (CBuf->flags & BUFFER_MODIFIED) star0 = star1 = '*';
    else star0 = star1 = '-';
@@ -835,7 +855,7 @@ static int screen_input_pending (int tsec)
  */
 static void mark_window_attributes (int wa)
 {
-   register Screen_Type *s = &JScreen[JWindow->top - 1],
+   register Screen_Type *s = &JScreen[JWindow->sy],
      *smax = s + JWindow->rows, *s1, *s2;
    Mark *m;
    register Line *l = JWindow->beg.line, *ml, *cline;
@@ -865,7 +885,7 @@ static void mark_window_attributes (int wa)
    
    m = CBuf->marks;
    
-   while ((m->flags & VISIBLE_MARK) == 0) m = m->next;
+   while ((m->flags & VISIBLE_MARK_MASK) == 0) m = m->next;
    ml = m->line;
    mn = m->n;			       /* already in canonical form */
    pn = LineNum + CBuf->nup;	       /* not in canonical form */
@@ -1034,7 +1054,7 @@ static int update_1(Line *top, int force)
 {
    int i;
    Window_Type *w;
-   int did_eob = 0, time_has_expired = 0;
+   int did_eob, time_has_expired = 0;
    
    if (Batch ||
        (!force
@@ -1124,7 +1144,7 @@ static int update_1(Line *top, int force)
 	
 	did_eob = 0;
 	
-	i = JWindow->top - 1;
+	i = JWindow->sy;
 	imax = i + JWindow->rows;
 	
 	while (i < imax)
@@ -1153,18 +1173,16 @@ static int update_1(Line *top, int force)
 		  if (((top == NULL) || (top->len == 0))
 		      && (Want_Eob && !did_eob && !(CBuf->flags & READ_ONLY)))
 		    {
-		       display_line(&Eob_Line, i + 1);
+		       display_line(&Eob_Line, i, JWindow->sx);
 		       
 		       /* JScreen[i].line = top; */
 		       did_eob = 1;
 		    }
-		  else display_line(top, i + 1);
+		  else display_line(top, i, JWindow->sx);
 	       }
 	     
 	     if (top != NULL)
-	       {
-		  top = top->next;
-	       }
+	       top = top->next;
 	     i++;
 	  }
 	
@@ -1206,7 +1224,7 @@ static void update_minibuffer(void)
 	  (void) screen_input_pending (10);
 #endif
 	JWindow->beg.line = CLine;
-	display_line(CLine, Jed_Num_Screen_Rows);
+	display_line(CLine, Jed_Num_Screen_Rows-1, 0);
 	while (w != JWindow) other_window();
 	Mini_Ghost = 1;
      }
@@ -1219,7 +1237,7 @@ static void update_minibuffer(void)
    else Mini_Ghost = ((*Message_Buffer) || (*Error_Buffer));
    
    if (Mini_Ghost == 0)
-     display_line(NULL, Jed_Num_Screen_Rows);
+     display_line(NULL, Jed_Num_Screen_Rows-1, 0);
 }
 
 void do_dialog(char *b)
@@ -1269,7 +1287,7 @@ void do_dialog(char *b)
 
 static void set_hscroll(int col)
 {
-   int hdiff, whs = abs(Wants_HScroll), wc = JWindow->column - 1,
+   int hdiff, whs = abs(Wants_HScroll), wc = JWindow->hscroll_column - 1,
      sw = Jed_Num_Screen_Cols - 1;
    static Line *last;
    Line *tmp;
@@ -1324,7 +1342,7 @@ static void set_hscroll(int col)
 	
 	if (Wants_HScroll < 0)
 	  {
-	     JWindow->column += HScroll;
+	     JWindow->hscroll_column += HScroll;
 	     touch_window();
 	     HScroll = 0;
 	  }
@@ -1351,13 +1369,13 @@ static void update_top_screen_line (void)
 {
 #if JED_HAS_TTY_MENUS
    if (Jed_Menus_Active
-       || (Top_Window_Row != 1))
+       || (Top_Window_SY > 0))
      {
 	jed_redraw_menus ();
 	return;
      }
 #else
-   if (Top_Window_Row == 1)
+   if (Top_Window_SY == 0)
      return;
    
    SLsmg_gotorc (0,0);
@@ -1376,6 +1394,9 @@ void update(Line *line, int force, int flag, int run_update_hook)
    static unsigned long last_time;
    Line *hscroll_line_save;
    
+   if (0 == Jed_Display_Initialized)
+     return;
+
 #if JED_HAS_SUBPROCESSES
    if (Child_Status_Changed_Flag)
      {
@@ -1461,7 +1482,7 @@ void update(Line *line, int force, int flag, int run_update_hook)
 #if JED_HAS_TTY_MENUS
    update_top_screen_line ();
 #else
-   if ((Top_Window_Row != 1) && JScreen[0].is_modified)
+   if ((Top_Window_SY > 0) && JScreen[0].is_modified)
      {
 	update_top_screen_line ();
      }
@@ -1511,7 +1532,7 @@ void register_change(int n)
    
    do
      {
-	s = &JScreen[w->top - 1];
+	s = &JScreen[w->sy];
 	smax = s + w->rows;
 	
 	while (s < smax)
@@ -1552,8 +1573,6 @@ int jed_get_screen_size (int *r, int *c)
    return 0;
 }
 
-static int Display_Initialized;
-
 static void dealloc_display (void)
 {
    if (JScreen != NULL)
@@ -1565,13 +1584,13 @@ static void dealloc_display (void)
 
 void jed_reset_display (void)
 {
-   if (Display_Initialized == 0) return;
+   if (Jed_Display_Initialized == 0) return;
    if (Batch) return;
    
-   Display_Initialized = 0;
+   Jed_Display_Initialized = 0;
    
 #if defined(VMS) || defined(REAL_UNIX_SYSTEM)
-   (void) SLang_run_hooks ("reset_display_hook", 0);
+   (void) jed_va_run_hooks ("_jed_reset_display_hooks", JED_HOOKS_RUN_ALL, 0);
 #endif
    
    dealloc_display ();
@@ -1586,10 +1605,9 @@ static void alloc_display (void)
    jed_get_screen_size (&r, &c);
    jed_update_window_sizes (r, c);
    
-   if (NULL == (JScreen = (Screen_Type *) SLmalloc (sizeof (Screen_Type) * r)))
+   if (NULL == (JScreen = (Screen_Type *) jed_malloc0 (sizeof (Screen_Type) * r)))
      exit_error ("Out of memory", 0);
    
-   memset ((char *) JScreen, 0, r * sizeof(Screen_Type));
    for (i = 0; i < r; i++)
      JScreen[i].is_modified = 1;
    
@@ -1609,16 +1627,16 @@ void jed_init_display (void)
    
    
 #if defined(VMS) || defined(REAL_UNIX_SYSTEM)
-   (void) SLang_run_hooks ("init_display_hook", 0);
+   (void) jed_va_run_hooks ("_jed_init_display_hooks", JED_HOOKS_RUN_ALL, 0);
 #endif
-   Display_Initialized = 1;
+   Jed_Display_Initialized = 1;
 }
 
 void jed_resize_display (void)
 {
    Jed_Resize_Pending = 0;
    
-   if (Display_Initialized == 0)
+   if (Jed_Display_Initialized == 0)
      return;
    
    dealloc_display ();
@@ -1661,7 +1679,7 @@ void jed_redraw_screen (int force)
      {
 	for (row = 0; row < JWindow->rows; row++)
 	  {
-	     JScreen[row + JWindow->top - 1].line = NULL;
+	     JScreen[row + JWindow->sy].line = NULL;
 	  }
 	l = NULL;
      }
@@ -1674,7 +1692,10 @@ void recenter(int *np)
 {
    Line *l = CLine;
    int i, n = *np;
-   
+
+   if (Batch)
+     return;
+
    JWindow->trashed = 1;
    if (n == 0)
      {
@@ -1714,7 +1735,9 @@ void recenter(int *np)
 	n--;
      }
    
-   update(l, 1, 0, 1);
+   /* update(l, 1, 0, 1); */
+   JScreen [JWindow->sy].line = l;
+   JScreen [JWindow->sy].is_modified = 1;
 }
 
 int window_line (void)
@@ -1756,7 +1779,7 @@ void touch_window (void)
    if (JScreen == NULL)
      return;
    
-   s = JScreen + (JWindow->top - 1);
+   s = JScreen + (JWindow->sy);
    if (JWindow->rows > Jed_Num_Screen_Rows)
      smax = s + Jed_Num_Screen_Rows;
    else
@@ -1786,7 +1809,7 @@ void touch_screen (void)
 	JWindow = JWindow->next;
      }
    while(w != JWindow);
-   if (Top_Window_Row != 1) JScreen[0].is_modified = 1;
+   if (Top_Window_SY > 0) JScreen[0].is_modified = 1;
 }
 
 void exit_error(char *str, int severity)
@@ -1798,11 +1821,14 @@ void exit_error(char *str, int severity)
    
    SLang_Error = SLKeyBoard_Quit = 0;
    auto_save_all();
+#if JED_HAS_EMACS_LOCKING
+   (void) jed_unlock_buffer_files ();
+#endif
    jed_reset_display();
    reset_tty();
-   fprintf(stderr,"\
-\007\rJED (%s): Fatal Error: %s\nS-Lang Version: %u\n",
-	   JED_VERSION, str, SLANG_VERSION);
+   
+   fprintf (stderr, "***Fatal Error: %s\n\n", str);
+   jed_show_version (stderr);
    
    if (*Error_Buffer) fprintf (stderr, "%s\n", Error_Buffer);
    if (CBuf != NULL)
@@ -1814,7 +1840,7 @@ void exit_error(char *str, int severity)
 					CLine->data, CLine->len, CLine->next, CLine->prev);
 	     fprintf(stderr, "Max_LineNum: %d, LineNum: %d\n", Max_LineNum, LineNum);
 	     if (JWindow != NULL) fprintf(stderr, "JWindow: %p, top: %d, rows: %d, buffer: %p\n",
-					  JWindow, JWindow->top, JWindow->rows, JWindow->buffer);
+					  JWindow, JWindow->sy, JWindow->rows, JWindow->buffer);
 	  }
      }
    if (severity)
@@ -1838,7 +1864,7 @@ void touch_window_hard(Window_Type *w, int all)
      {
 	Screen_Type *s, *smax;
 	
-	s = JScreen + (w->top - 1);
+	s = JScreen + (w->sy);
 	if (w->rows > Jed_Num_Screen_Rows)
 	  smax = s + Jed_Num_Screen_Rows;
 	else
@@ -1856,7 +1882,8 @@ void touch_window_hard(Window_Type *w, int all)
    while (all && (w != wsave));
 }
 
-int jed_find_line_on_screen (Line *l)
+#if 1
+int jed_find_line_on_screen (Line *l, int min_line)
 {
    int i, imax;
    
@@ -1864,14 +1891,14 @@ int jed_find_line_on_screen (Line *l)
      return -1;
    
    imax = Jed_Num_Screen_Rows - 2;
-   for (i = 0; i < imax; i++)
+   for (i = min_line; i < imax; i++)
      {
 	if (JScreen[i].line == l)
 	  return i;
      }
    return -1;
 }
-
+#endif
 
 void update_cmd (int *force)
 {
