@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -8,6 +8,7 @@
 #define __DAVIS_SYSDEP_H__
 
 extern int Ignore_User_Abort;
+extern char *_Jed_Backspace_Key;
 
 #ifdef __MSDOS_16BIT__
 #define MAX_INPUT_BUFFER_LEN 40
@@ -48,14 +49,13 @@ extern int sys_findnext(char *);
 extern unsigned long  sys_time(void);
 extern int Meta_Char;
 extern int DEC_8Bit_Hack;
-extern int Abort_Char;
 extern void map_character(int *, int *);
 extern int make_random_number(int *, int *);
 extern int ch_dir(char *);
 extern void deslash(char *);
 extern char *slash2slash (char *);
 
-#if defined(__MSDOS_16BIT__) || defined(__os2__)
+#if defined(__MSDOS_16BIT__) || defined(__os2__) || defined(__WIN32__)
 extern int NumLock_Is_Gold;
 #endif
 
@@ -63,7 +63,7 @@ extern int NumLock_Is_Gold;
 extern int IsHPFSFileSystem(char *);
 #endif
 
-#if defined(__GO32__) || defined(__WATCOMC__)
+#if defined(__GO32__) || (defined(__WATCOMC__) && !defined(__QNX__))
 extern void i386_check_kbd(void);
 extern int i386_access (char *, int);
 #endif
@@ -77,6 +77,7 @@ extern int sys_System (char *);
 # ifdef MSWINDOWS
 extern int PC_Fn_Char;
 # endif
+extern int Jed_Filename_Case_Sensitive;
 #else
 # define SCREEN_HEIGHT 24
 #endif
@@ -104,12 +105,6 @@ extern int jed_win32_rename (char *, char *);
 #ifdef VMS
 #endif
 
-#ifdef __unix__
-#ifndef __GO32__
-extern void get_passwd_info (char *);
-#endif
-#endif
-
 extern char *get_cwd(void);
 
 extern void jed_pause (int *);
@@ -117,4 +112,7 @@ extern void sys_pause (int);
 
 extern char Jed_Root_Dir [JED_MAX_PATH_LEN];
 
+extern void (*Jed_Sig_Exit_Fun) (void);
+
+
 #endif /* __DAVIS_SYSDEP_H__ */
